/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.demo.simpleapp;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CountDownLatch;

import org.openmuc.framework.data.DoubleValue;
import org.openmuc.framework.data.Flag;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.dataaccess.Channel;
import org.openmuc.framework.dataaccess.DataAccessService;
import org.openmuc.framework.dataaccess.LogSourceNotAvailableException;
import org.openmuc.framework.dataaccess.RecordListener;
import org.openmuc.framework.dataaccess.WriteValueContainer;
import org.osgi.service.component.ComponentContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class SimpleDemoApp extends Thread implements RecordListener {

	private final static Logger logger = LoggerFactory.getLogger(SimpleDemoApp.class);

	private volatile CountDownLatch threadFinishedSignal;
	private volatile boolean deactivatedSignal;

	DataAccessService dataAccessService;
	Channel channel1, channel2;
	public static final String channelLabel1 = "channel1";
	public static final String channelLabel2 = "channel2";

	protected void activate(ComponentContext context) {
		logger.info("Activating Demo App");
		start();

	}

	protected void deactivate(ComponentContext context) {
		logger.info("Deactivating Demo App");
		threadFinishedSignal = new CountDownLatch(1);
		deactivatedSignal = true;
		interrupt();
		try {
			threadFinishedSignal.await();
		} catch (InterruptedException e) {
		}
	}

	protected void setDataAccessService(DataAccessService dataAccessService) {
		this.dataAccessService = dataAccessService;
	}

	protected void unsetDataAccessService(DataAccessService dataAccessService) {
		this.dataAccessService = null;
	}

	@Override
	public void run() {

		logger.info("Demo App started running...");

		if (deactivatedSignal) {
			logger.info("DemoApp thread interrupted: will stop");
			threadFinishedSignal.countDown();
			return;
		}

		channel1 = dataAccessService.getChannel(channelLabel1);
		channel2 = dataAccessService.getChannel(channelLabel2);

		while (channel1 == null || channel2 == null) {
			String errorMessage = "Channel " + channelLabel1 + " or " + channelLabel2
					+ " not found, will try again in 5 seconds.";
			logger.error(errorMessage);

			try {
				Thread.sleep(5000);
			} catch (InterruptedException e) {
				logger.info("DemoApp thread interrupted: will stop");
				threadFinishedSignal.countDown();
				return;
			}

			channel1 = dataAccessService.getChannel(channelLabel1);
			channel2 = dataAccessService.getChannel(channelLabel2);
		}

		channel1.addListener(this);

		WriteValueContainer writeContainer = channel1.getWriteContainer();
		writeContainer.setValue(new DoubleValue(23.0));

		List<WriteValueContainer> writeContainerList = new ArrayList<WriteValueContainer>(1);
		writeContainerList.add(writeContainer);

		while (true) {

			if (deactivatedSignal) {
				logger.info("DemoApp thread interrupted: will stop");
				threadFinishedSignal.countDown();
				return;
			}

			List<Record> records = null;
			try {
				records = channel1.getLoggedValues(System.currentTimeMillis() - 50000);
			} catch (LogSourceNotAvailableException e1) {
				logger.error("Unable to get logged values: log source unavailable");
			} catch (IOException e1) {
				logger.error("Unable to get logged values: IOException: " + e1.getMessage());
			}
			if (records != null) {
				for (Record record : records) {
					logger.info("Got logged record: " + record);
				}
			}

			Flag flag = channel1.write(new DoubleValue(33.0), 3000);
			logger.info("Wrote to channel1, resulting flag was: " + flag);

			logger.info("Channel1 - latest sample: " + channel1.getLatestSample());
			logger.info("Channel1 - channel state: " + channel1.getChannelState());
			logger.info("Channel1 - device state: " + channel1.getDeviceState());

			logger.info("Channel2 - latest sample: " + channel2.getLatestSample());
			logger.info("Channel2 - channel state: " + channel2.getChannelState());
			logger.info("Channel2 - device state: " + channel2.getDeviceState());

			try {
				Thread.sleep(4000);
			} catch (InterruptedException e) {
				logger.info("DemoApp thread interrupted: will stop");
				channel1.removeListener(this);
				threadFinishedSignal.countDown();
				return;
			}
		}

	}

	@Override
	public void newRecord(Channel channel, Record record) {
		logger.info("Record listener got new record for channel: " + channel.getLabel() + " : " + record);
	}
}
