/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.core.datamanager;

import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CountDownLatch;

import org.openmuc.framework.config.ChannelConfig;
import org.openmuc.framework.data.DoubleValue;
import org.openmuc.framework.data.Flag;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.data.TypeConversionException;
import org.openmuc.framework.data.Value;
import org.openmuc.framework.data.ValueType;
import org.openmuc.framework.dataaccess.Channel;
import org.openmuc.framework.dataaccess.ChannelState;
import org.openmuc.framework.dataaccess.DeviceState;
import org.openmuc.framework.dataaccess.LogSourceNotAvailableException;
import org.openmuc.framework.dataaccess.ReadRecordContainer;
import org.openmuc.framework.dataaccess.RecordListener;
import org.openmuc.framework.dataaccess.WriteValueContainer;
import org.openmuc.framework.datalogger.spi.LogChannel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class ChannelImpl implements Channel {

	private final static Logger logger = LoggerFactory.getLogger(ChannelImpl.class);

	volatile Record currentSample;
	ChannelRecordContainerImpl driverChannel;
	volatile ChannelConfigImpl config;
	ChannelCollection samplingCollection;
	ChannelCollection loggingCollection;
	private final Set<RecordListener> listeners = new LinkedHashSet<RecordListener>();
	private final DataManager dataManager;
	Object handle;

	public ChannelImpl(DataManager dataManager, ChannelConfigImpl config, ChannelState initState, Flag initFlag,
			long currentTime, List<LogChannel> logChannels) {
		this.dataManager = dataManager;
		this.config = config;

		if (config.disabled) {
			config.state = ChannelState.DISABLED;
			currentSample = new Record(null, null, Flag.DISABLED);
		}
		else if (!config.isListening() && config.samplingInterval < 0) {
			config.state = initState;
			currentSample = new Record(null, null, Flag.SAMPLING_AND_LISTENING_DISABLED);
		}
		else {
			config.state = initState;
			currentSample = new Record(null, null, initFlag);
		}

		if (config.loggingInterval != ChannelConfig.LOGGING_INTERVAL_DEFAULT) {
			dataManager.addToLoggingCollections(this, currentTime);
			logChannels.add(config);
		}
	}

	@Override
	public String getLabel() {
		return config.label;
	}

	@Override
	public String getChannelAddress() {
		return config.channelAddress;
	}

	@Override
	public String getDescription() {
		return config.description;
	}

	@Override
	public String getUnit() {
		return config.unit;
	}

	@Override
	public ValueType getValueType() {
		return config.valueType;
	}

	@Override
	public int getSamplingInterval() {
		return config.samplingInterval;
	}

	@Override
	public int getSamplingTimeOffset() {
		return config.samplingTimeOffset;
	}

	@Override
	public int getLoggingInterval() {
		return config.loggingInterval;
	}

	@Override
	public int getLoggingTimeOffset() {
		return config.loggingTimeOffset;
	}

	@Override
	public String getDriverName() {
		return config.deviceParent.driverParent.id;
	}

	@Override
	public String getInterfaceAddress() {
		return config.deviceParent.interfaceAddress;
	}

	@Override
	public String getDeviceAddress() {
		return config.deviceParent.deviceAddress;
	}

	@Override
	public String getDeviceName() {
		return config.deviceParent.name;
	}

	@Override
	public String getDeviceDescription() {
		return config.deviceParent.description;
	}

	@Override
	public ChannelState getChannelState() {
		return config.state;
	}

	@Override
	public DeviceState getDeviceState() {
		return config.deviceParent.device.state;
	}

	@Override
	public void addListener(RecordListener listener) {
		synchronized (listeners) {
			listeners.add(listener);
		}
	}

	@Override
	public void removeListener(RecordListener listener) {
		synchronized (listeners) {
			listeners.remove(listener);
		}
	}

	@Override
	public Record getLatestSample() {
		return currentSample;
	}

	@Override
	public void setLatestSample(Record record) {
		setNewSample(record);
	}

	@Override
	public Record getLoggedValue(long timestamp) throws LogSourceNotAvailableException, IOException {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public List<Record> getLoggedValues(long startTime) throws LogSourceNotAvailableException, IOException {
		synchronized (dataManager.activeLogSources) {
			if (dataManager.activeLogSources.size() == 0) {
				throw new LogSourceNotAvailableException();
			}
			return dataManager.activeLogSources.get(0).getValues(config.id, startTime);
		}
	}

	@Override
	public List<Record> getLoggedValues(long startTime, long endTime) throws LogSourceNotAvailableException,
			IOException {
		synchronized (dataManager.activeLogSources) {
			if (dataManager.activeLogSources.size() == 0) {
				throw new LogSourceNotAvailableException();
			}
			return dataManager.activeLogSources.get(0).getValues(config.id, startTime, endTime);
		}
	}

	void setNewSample(Record record) {

		if (record.getFlag() == Flag.VALID) {
			Double scalingFactor = config.scalingFactor;
			Double scalingOffset = config.valueOffset;

			try {
				if (scalingFactor != null) {
					record = new Record(new DoubleValue(record.getValue().asDouble() * scalingFactor),
							record.getTimestamp(), record.getFlag());
				}
				if (scalingOffset != null) {
					record = new Record(new DoubleValue(record.getValue().asDouble() + scalingOffset),
							record.getTimestamp(), record.getFlag());
				}
			} catch (TypeConversionException e) {
				logger.error("Unable to apply scaling factor because a TypeConversionError occured. ByteStrings cannot be scaled!");
			}

			currentSample = record;
		}
		else {
			currentSample = new Record(currentSample.getValue(), currentSample.getTimestamp(), record.getFlag());
		}
		notifyListeners();
	}

	private void notifyListeners() {
		if (listeners.size() != 0) {
			synchronized (listeners) {
				for (RecordListener listener : listeners) {
					config.deviceParent.device.dataManager.executor.execute(new ListenerNotifier(listener, this,
							currentSample));
				}
			}
		}
	}

	ChannelRecordContainerImpl createChannelRecordContainer() {
		return new ChannelRecordContainerImpl(this);
	}

	void setFlag(Flag flag) {
		if (flag != currentSample.getFlag()) {
			currentSample = new Record(currentSample.getValue(), currentSample.getTimestamp(), flag);
			notifyListeners();
		}
	}

	public void setNewDeviceState(ChannelState state, Flag flag) {
		if (config.disabled) {
			config.state = ChannelState.DISABLED;
			setFlag(Flag.DISABLED);
		}
		else if (!config.isListening() && config.samplingInterval < 0) {
			config.state = state;
			setFlag(Flag.SAMPLING_AND_LISTENING_DISABLED);
		}
		else {
			config.state = state;
			setFlag(flag);
		}
	}

	@Override
	public Flag write(Value value, int timeout) {
		CountDownLatch writeTaskFinishedSignal = new CountDownLatch(1);
		WriteValueContainerImpl writeValueContainer = new WriteValueContainerImpl(this);

		Value adjustedValue = value;

		Double valueOffset = config.valueOffset;
		Double scalingFactor = config.scalingFactor;

		if (valueOffset != null) {
			adjustedValue = new DoubleValue(adjustedValue.asDouble() - valueOffset);
		}
		if (scalingFactor != null) {
			adjustedValue = new DoubleValue(adjustedValue.asDouble() / scalingFactor);
		}
		writeValueContainer.setValue(adjustedValue);

		List<WriteValueContainerImpl> writeValueContainerList = new ArrayList<WriteValueContainerImpl>(1);
		writeValueContainerList.add(writeValueContainer);
		WriteTask writeTask = new WriteTask(dataManager, config.deviceParent.device, writeValueContainerList, timeout,
				writeTaskFinishedSignal);
		synchronized (dataManager.newWriteTasks) {
			dataManager.newWriteTasks.add(writeTask);
		}
		dataManager.interrupt();
		try {
			writeTaskFinishedSignal.await();
		} catch (InterruptedException e) {
		}

		if (writeValueContainer.getFlag() == Flag.VALID) {
			setNewSample(new Record(value, System.currentTimeMillis()));
		}

		return writeValueContainer.getFlag();
	}

	@Override
	public Record read(int timeout) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public boolean isConnected() {
		if (config.state == ChannelState.CONNECTED || config.state == ChannelState.SAMPLING
				|| config.state == ChannelState.LISTENING) {
			return true;
		}
		return false;
	}

	@Override
	public WriteValueContainer getWriteContainer() {
		return new WriteValueContainerImpl(this);
	}

	@Override
	public ReadRecordContainer getReadContainer() {
		// TODO Auto-generated method stub
		return null;
	}

}
