/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.core.datamanager;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map.Entry;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.locks.ReentrantLock;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;

import org.openmuc.framework.config.ArgumentSyntaxException;
import org.openmuc.framework.config.ChannelScanInformation;
import org.openmuc.framework.config.ConfigChangeListener;
import org.openmuc.framework.config.ConfigService;
import org.openmuc.framework.config.ConfigWriteException;
import org.openmuc.framework.config.DeviceScanInformation;
import org.openmuc.framework.config.DriverChangeListener;
import org.openmuc.framework.config.DriverNotAvailableException;
import org.openmuc.framework.config.ParseException;
import org.openmuc.framework.config.RootConfig;
import org.openmuc.framework.data.Flag;
import org.openmuc.framework.dataaccess.Channel;
import org.openmuc.framework.dataaccess.ChannelChangeListener;
import org.openmuc.framework.dataaccess.ChannelState;
import org.openmuc.framework.dataaccess.DataAccessService;
import org.openmuc.framework.dataaccess.DeviceState;
import org.openmuc.framework.dataaccess.LogicalDevice;
import org.openmuc.framework.dataaccess.LogicalDeviceChangeListener;
import org.openmuc.framework.dataaccess.WriteValueContainer;
import org.openmuc.framework.datalogger.spi.LogChannel;
import org.openmuc.framework.datalogger.spi.LogRecordContainer;
import org.openmuc.framework.datalogger.spi.LogSinkService;
import org.openmuc.framework.datalogger.spi.LogSourceService;
import org.openmuc.framework.driver.spi.ChannelRecordContainer;
import org.openmuc.framework.driver.spi.ConnectionException;
import org.openmuc.framework.driver.spi.DriverService;
import org.openmuc.framework.driver.spi.RecordsReceivedListener;
import org.osgi.service.component.ComponentContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class DataManager extends Thread implements DataAccessService, ConfigService, RecordsReceivedListener {

	private final static Logger logger = LoggerFactory.getLogger(DataManager.class);

	private volatile boolean stopFlag = false;

	private final HashMap<String, DriverService> newDrivers = new LinkedHashMap<String, DriverService>();;
	// does not need to be a list because RemovedService() for driver services
	// are never called in parallel:
	private volatile String driverToBeRemovedId = null;
	private volatile LogSinkService logSinkToBeRemoved = null;
	private volatile LogSourceService logSourceToBeRemoved = null;

	final List<Device> connected = new LinkedList<Device>();
	final List<Device> disconnected = new LinkedList<Device>();
	final List<Device> connectionFailures = new LinkedList<Device>();
	final List<SamplingTask> samplingTaskFinished = new LinkedList<SamplingTask>();
	final List<WriteTask> newWriteTasks = new LinkedList<WriteTask>();
	final List<DeviceTask> tasksFinished = new LinkedList<DeviceTask>();
	private volatile RootConfigImpl newRootConfigWithoutDefaults = null;
	private final HashMap<String, DriverService> activeDrivers = new LinkedHashMap<String, DriverService>();

	private CountDownLatch dataManagerStoppedSignal;

	private final List<Action> actions = new LinkedList<Action>();
	private final List<ConfigChangeListener> configChangeListeners = new LinkedList<ConfigChangeListener>();

	CountDownLatch logSinkRemovedSignal;
	private final List<LogSinkService> newLogSinks = new LinkedList<LogSinkService>();
	private final List<LogSinkService> activeLogSinks = new LinkedList<LogSinkService>();

	CountDownLatch logSourceRemovedSignal;
	private final List<LogSourceService> newLogSources = new LinkedList<LogSourceService>();
	final List<LogSourceService> activeLogSources = new LinkedList<LogSourceService>();

	private final List<List<ChannelRecordContainer>> receivedRecordContainers = new LinkedList<List<ChannelRecordContainer>>();

	volatile int activeDeviceCountDown;

	private volatile RootConfigImpl rootConfig;
	private volatile RootConfigImpl rootConfigWithoutDefaults;

	private File configFile;

	ExecutorService executor = null;

	private volatile Boolean dataManagerStarted = false;

	CountDownLatch driverRemovedSignal;

	private CountDownLatch newConfigSignal;

	private final ReentrantLock configLock = new ReentrantLock();

	protected void activate(ComponentContext context) throws ParseException, TransformerFactoryConfigurationError,
			IOException, ParserConfigurationException, TransformerException {
		logger.info("Activating Data Manager");

		executor = Executors.newCachedThreadPool();

		String configFileName = System.getProperty("org.openmuc.configfile");
		if (configFileName == null) {
			configFileName = "openmuc-config.xml";
		}
		configFile = new File(configFileName);

		try {
			rootConfigWithoutDefaults = RootConfigImpl.createFromFile(configFile);
		} catch (FileNotFoundException e) {
			// create an empty configuration and store it in a file
			rootConfigWithoutDefaults = new RootConfigImpl();
			rootConfigWithoutDefaults.writeToFile(configFile);
			logger.info("No configuration file found. Created an empty config file at: " + configFile.getAbsolutePath());
		}

		rootConfig = new RootConfigImpl();

		applyConfiguration(rootConfigWithoutDefaults, System.currentTimeMillis());

		setName("OpenMUC Datamanager");
		start();

		dataManagerStarted = true;

	}

	protected void deactivate(ComponentContext context) {
		logger.info("Deactivating Data Manager");
		stopFlag = true;

		dataManagerStoppedSignal = new CountDownLatch(1);
		interrupt();
		try {
			dataManagerStoppedSignal.await();
		} catch (InterruptedException e) {
		}
		dataManagerStarted = false;
	}

	@Override
	public void run() {

		if (stopFlag) {
			prepareStop();
			return;
		}
		handleInterruptEvent();

		while (true) {

			if (actions.size() == 0) {
				try {
					while (true) {
						Thread.sleep(Long.MAX_VALUE);
					}
				} catch (InterruptedException e) {
					if (stopFlag) {
						prepareStop();
						return;
					}
					handleInterruptEvent();
					continue;
				}
			}

			long sleepTime = 0;
			Action currentAction = actions.get(0);

			sleepTime = currentAction.startTime - System.currentTimeMillis();
			if (sleepTime > 0) {
				try {
					Thread.sleep(sleepTime);
				} catch (InterruptedException e1) {
					if (stopFlag) {
						prepareStop();
						return;
					}
					handleInterruptEvent();
					continue;
				}
			}
			else {
				if (interrupted()) {
					if (stopFlag) {
						prepareStop();
						return;
					}
					handleInterruptEvent();
					continue;
				}
			}
			actions.remove(0);

			if (currentAction.timeouts != null && currentAction.timeouts.size() > 0) {
				for (SamplingTask samplingTask : currentAction.timeouts) {
					samplingTask.timeout();
				}
			}

			if (currentAction.loggingCollections != null && currentAction.loggingCollections.size() > 0) {

				List<LogRecordContainer> logContainers = new LinkedList<LogRecordContainer>();
				List<ChannelImpl> toRemove = new LinkedList<ChannelImpl>();

				for (ChannelCollection loggingCollection : currentAction.loggingCollections) {
					toRemove.clear();
					for (ChannelImpl channel : loggingCollection.channels) {
						if (channel.config.state == ChannelState.DELETED) {
							toRemove.add(channel);
						}
						else if (!channel.config.isDisabled()) {
							logContainers.add(new LogRecordContainerImpl(channel.config.id, channel.currentSample));
						}
					}

					for (ChannelImpl channel : toRemove) {
						loggingCollection.channels.remove(channel);
					}

					if (loggingCollection.channels.size() > 0) {
						addLoggingCollectionToActions(loggingCollection, currentAction.startTime
								+ loggingCollection.interval);
					}
				}
				for (LogSinkService logSink : activeLogSinks) {
					logSink.log(logContainers, currentAction.startTime);
				}
			}

			if (currentAction.connectionRetryDevices != null && currentAction.connectionRetryDevices.size() > 0) {
				for (Device device : currentAction.connectionRetryDevices) {
					device.connectRetrySignal();
				}
			}

			if (currentAction.samplingCollections != null && currentAction.samplingCollections.size() > 0) {

				// List<ChannelImpl> toRemove = new LinkedList<ChannelImpl>();
				// HashMap<String, HashMap<Device, List<ChannelRecordContainerImpl>>> channels = new HashMap<String,
				// HashMap<Device, List<ChannelRecordContainerImpl>>>();
				// for (ChannelCollection samplingCollection : currentAction.samplingCollections) {
				//
				// toRemove.clear();
				//
				// for (ChannelImpl channel : samplingCollection.channels) {
				// HashMap<Device, List<ChannelRecordContainerImpl>> selectedDriver = channels
				// .get(channel.config.deviceParent.driverParent.id);
				// if (selectedDriver == null) {
				// selectedDriver = new HashMap<Device, List<ChannelRecordContainerImpl>>();
				// channels.put(channel.config.deviceParent.driverParent.id, selectedDriver);
				// }
				// List<ChannelRecordContainerImpl> selectedDevice = selectedDriver
				// .get(channel.config.deviceParent.device);
				// if (selectedDevice == null) {
				// selectedDevice = new ArrayList<ChannelRecordContainerImpl>();
				// selectedDriver.put(channel.config.deviceParent.device, selectedDevice);
				// }
				//
				// if (!channel.config.isDisabled()) {
				// selectedDevice.add(channel.createChannelRecordContainer());
				// }
				// else {
				// toRemove.add(channel);
				// }
				// }
				//
				// for (ChannelImpl channel : toRemove) {
				// samplingCollection.channels.remove(channel);
				// }
				//
				// if (samplingCollection.channels.size() > 0) {
				// addSamplingCollectionToActions(samplingCollection, currentAction.startTime
				// + samplingCollection.interval);
				// }
				// }
				//
				// for (HashMap<Device, List<ChannelRecordContainerImpl>> channelsByDevice : channels.values()) {
				// for (Entry<Device, List<ChannelRecordContainerImpl>> selectedChannels : channelsByDevice.entrySet())
				// {
				// Device device = selectedChannels.getKey();
				//
				// SamplingTask samplingTask = new SamplingTask(this, selectedChannels.getKey(),
				// selectedChannels.getValue(), device.deviceConfig.readTimeout);
				//
				// device.addSampleTask(samplingTask);
				//
				// addReadWorkerTimeoutToActions(samplingTask, currentAction.startTime
				// + device.deviceConfig.readTimeout);
				// }
				// }

				for (ChannelCollection samplingCollection : currentAction.samplingCollections) {
					List<ChannelRecordContainerImpl> selectedChannels = new ArrayList<ChannelRecordContainerImpl>(
							samplingCollection.channels.size());
					for (ChannelImpl channel : samplingCollection.channels) {
						selectedChannels.add(channel.createChannelRecordContainer());
					}
					SamplingTask samplingTask = new SamplingTask(this, samplingCollection.device, selectedChannels,
							samplingCollection.device.deviceConfig.readTimeout, samplingCollection.samplingGroup);

					samplingCollection.device.addSampleTask(samplingTask);

					addReadWorkerTimeoutToActions(samplingTask, currentAction.startTime
							+ samplingCollection.device.deviceConfig.readTimeout);

					addSamplingCollectionToActions(samplingCollection, currentAction.startTime
							+ samplingCollection.interval);
				}

			}
		}
	}

	private void addSamplingCollectionToActions(ChannelCollection channelCollection, long startTimestamp) {

		Action fittingAction = null;

		ListIterator<Action> actionIterator = actions.listIterator();
		while (actionIterator.hasNext()) {
			Action currentAction = actionIterator.next();
			if (currentAction.startTime == startTimestamp) {
				fittingAction = currentAction;
				if (fittingAction.samplingCollections == null) {
					fittingAction.samplingCollections = new LinkedList<ChannelCollection>();
				}
				break;
			}
			else if (currentAction.startTime > startTimestamp) {
				fittingAction = new Action(startTimestamp);
				fittingAction.samplingCollections = new LinkedList<ChannelCollection>();
				actionIterator.previous();
				actionIterator.add(fittingAction);
				break;
			}
		}

		if (fittingAction == null) {
			fittingAction = new Action(startTimestamp);
			fittingAction.samplingCollections = new LinkedList<ChannelCollection>();
			actions.add(fittingAction);
		}

		fittingAction.samplingCollections.add(channelCollection);
		channelCollection.action = fittingAction;

	}

	private void addLoggingCollectionToActions(ChannelCollection channelCollection, long startTimestamp) {

		Action fittingAction = null;

		ListIterator<Action> actionIterator = actions.listIterator();
		while (actionIterator.hasNext()) {
			Action currentAction = actionIterator.next();
			if (currentAction.startTime == startTimestamp) {
				fittingAction = currentAction;
				if (fittingAction.loggingCollections == null) {
					fittingAction.loggingCollections = new LinkedList<ChannelCollection>();
				}
				break;
			}
			else if (currentAction.startTime > startTimestamp) {
				fittingAction = new Action(startTimestamp);
				fittingAction.loggingCollections = new LinkedList<ChannelCollection>();
				actionIterator.previous();
				actionIterator.add(fittingAction);
				break;
			}
		}

		if (fittingAction == null) {
			fittingAction = new Action(startTimestamp);
			fittingAction.loggingCollections = new LinkedList<ChannelCollection>();
			actions.add(fittingAction);
		}

		fittingAction.loggingCollections.add(channelCollection);
		channelCollection.action = fittingAction;
	}

	void addReconnectDeviceToActions(Device device, long startTimestamp) {
		Action fittingAction = null;

		ListIterator<Action> actionIterator = actions.listIterator();
		while (actionIterator.hasNext()) {
			Action currentAction = actionIterator.next();
			if (currentAction.startTime == startTimestamp) {
				fittingAction = currentAction;
				if (fittingAction.connectionRetryDevices == null) {
					fittingAction.connectionRetryDevices = new LinkedList<Device>();
				}
				break;
			}
			else if (currentAction.startTime > startTimestamp) {
				fittingAction = new Action(startTimestamp);
				fittingAction.connectionRetryDevices = new LinkedList<Device>();
				actionIterator.previous();
				actionIterator.add(fittingAction);
				break;
			}
		}

		if (fittingAction == null) {
			fittingAction = new Action(startTimestamp);
			fittingAction.connectionRetryDevices = new LinkedList<Device>();
			actions.add(fittingAction);
		}

		fittingAction.connectionRetryDevices.add(device);
	}

	private void addReadWorkerTimeoutToActions(SamplingTask readWorker, long timeout) {

		Action fittingAction = null;

		ListIterator<Action> actionIterator = actions.listIterator();
		while (actionIterator.hasNext()) {
			Action currentAction = actionIterator.next();
			if (currentAction.startTime == timeout) {
				fittingAction = currentAction;
				if (fittingAction.timeouts == null) {
					fittingAction.timeouts = new LinkedList<SamplingTask>();
				}
				break;
			}
			else if (currentAction.startTime > timeout) {
				fittingAction = new Action(timeout);
				fittingAction.timeouts = new LinkedList<SamplingTask>();
				actionIterator.previous();
				actionIterator.add(fittingAction);
				break;
			}
		}

		if (fittingAction == null) {
			fittingAction = new Action(timeout);
			fittingAction.timeouts = new LinkedList<SamplingTask>();
			actions.add(fittingAction);
		}

		fittingAction.timeouts.add(readWorker);
	}

	private void handleInterruptEvent() {

		long currentTime = 0;

		if (newRootConfigWithoutDefaults != null) {
			currentTime = System.currentTimeMillis();
			applyConfiguration(newRootConfigWithoutDefaults, currentTime);
			newRootConfigWithoutDefaults = null;
			newConfigSignal.countDown();
		}

		synchronized (receivedRecordContainers) {
			if (receivedRecordContainers.size() != 0) {
				for (List<ChannelRecordContainer> recordContainers : receivedRecordContainers) {
					for (ChannelRecordContainer container : recordContainers) {
						ChannelRecordContainerImpl containerImpl = (ChannelRecordContainerImpl) container;
						if (containerImpl.channel.config.state == ChannelState.LISTENING) {
							containerImpl.channel.setNewSample(containerImpl.record);
						}
					}
				}
			}
			receivedRecordContainers.clear();
		}

		synchronized (samplingTaskFinished) {
			if (samplingTaskFinished.size() != 0) {
				for (SamplingTask samplingTask : samplingTaskFinished) {
					samplingTask.storeValues();
					samplingTask.device.taskFinished();
				}
				samplingTaskFinished.clear();
			}
		}

		synchronized (tasksFinished) {
			if (tasksFinished.size() != 0) {
				for (DeviceTask deviceTask : tasksFinished) {
					deviceTask.device.taskFinished();
				}
				tasksFinished.clear();
			}
		}

		synchronized (newDrivers) {
			if (newDrivers.size() != 0) {

				activeDrivers.putAll(newDrivers);
				for (Entry<String, DriverService> newDriverEntry : newDrivers.entrySet()) {
					logger.info("New driver registered: " + newDriverEntry.getKey());
					DriverConfigImpl driverConfig = rootConfig.driverConfigsById.get(newDriverEntry.getKey());
					if (driverConfig != null) {
						driverConfig.activeDriver = newDriverEntry.getValue();
						for (DeviceConfigImpl deviceConfig : driverConfig.deviceConfigsByAddress.values()) {
							deviceConfig.device.driverRegisteredSignal();
						}
					}
				}
				newDrivers.clear();
			}
		}

		synchronized (newLogSinks) {
			if (newLogSinks.size() != 0) {
				activeLogSinks.addAll(newLogSinks);
				for (LogSinkService logSink : newLogSinks) {
					logger.info("New log sink registered: " + logSink.getId());
					logSink.setChannelsToLog(rootConfig.logChannels);
				}
				newLogSinks.clear();
			}
		}

		synchronized (newLogSources) {
			if (newLogSources.size() != 0) {
				activeLogSources.addAll(newLogSources);
				logger.info("New log source(s) registerd");
				newLogSources.clear();
			}
		}

		if (driverToBeRemovedId != null) {
			if (activeDrivers.remove(driverToBeRemovedId) == null) {
				newDrivers.remove(driverToBeRemovedId);
				driverRemovedSignal.countDown();
			}
			else {
				logger.debug("---------------------------driver to be removed");
				DriverConfigImpl driverConfig = rootConfig.driverConfigsById.get(driverToBeRemovedId);

				if (driverConfig != null) {
					activeDeviceCountDown = driverConfig.deviceConfigsByAddress.size();
					for (DeviceConfigImpl deviceConfig : driverConfig.deviceConfigsByAddress.values()) {
						deviceConfig.device.driverDeregisteredSignal();
					}
					if (activeDeviceCountDown == 0) {
						driverRemovedSignal.countDown();
					}
				}
				else {
					driverRemovedSignal.countDown();
				}
			}
			driverToBeRemovedId = null;
		}

		if (logSinkToBeRemoved != null) {
			if (activeLogSinks.remove(logSinkToBeRemoved) == false) {
				newLogSinks.remove(logSinkToBeRemoved);
			}
			logSinkToBeRemoved = null;
			logSinkRemovedSignal.countDown();
		}

		if (logSourceToBeRemoved != null) {
			if (activeLogSources.remove(logSourceToBeRemoved) == false) {
				newLogSources.remove(logSourceToBeRemoved);
			}
			logSourceToBeRemoved = null;
			logSourceRemovedSignal.countDown();
		}

		synchronized (connectionFailures) {
			if (connectionFailures.size() != 0) {
				for (Device connectionFailureDevice : connectionFailures) {
					connectionFailureDevice.connectFailureSignal();
				}
				connectionFailures.clear();
			}
		}

		synchronized (connected) {
			if (connected.size() != 0) {
				for (Device connectedDevice : connected) {
					connectedDevice.connectedSignal(currentTime);
				}
				connected.clear();
			}
		}

		synchronized (newWriteTasks) {
			if (newWriteTasks.size() != 0) {
				for (WriteTask newWriteTask : newWriteTasks) {
					newWriteTask.device.addWriteTask(newWriteTask);
				}
				newWriteTasks.clear();
			}
		}

		synchronized (disconnected) {
			if (disconnected.size() != 0) {
				for (Device connectedDevice : disconnected) {
					connectedDevice.disconnectedSignal();
				}
				disconnected.clear();
			}
		}

	}

	private void applyConfiguration(RootConfigImpl configWithoutDefaults, long currentTime) {

		RootConfigImpl newRootConfig = configWithoutDefaults.cloneWithDefaults();

		List<LogChannel> logChannels = new LinkedList<LogChannel>();

		for (DriverConfigImpl oldDriverConfig : rootConfig.driverConfigsById.values()) {
			DriverConfigImpl newDriverConfig = newRootConfig.driverConfigsById.get(oldDriverConfig.id);
			if (newDriverConfig != null) {
				newDriverConfig.activeDriver = oldDriverConfig.activeDriver;
			}
			for (DeviceConfigImpl oldDeviceConfig : oldDriverConfig.deviceConfigsByAddress.values()) {
				DeviceConfigImpl newDeviceConfig = null;
				if (newDriverConfig != null) {
					newDeviceConfig = newDriverConfig.deviceConfigsByAddress.get(oldDeviceConfig.getIfAndDevAddress());
				}
				if (newDeviceConfig == null) {
					// Device was deleted in new config
					oldDeviceConfig.device.deleteSignal();
				}
				else {
					// Device exists in new and old config
					oldDeviceConfig.device.configChangedSignal(newDeviceConfig, currentTime, logChannels);
				}
			}
		}

		for (DriverConfigImpl newDriverConfig : newRootConfig.driverConfigsById.values()) {
			DriverConfigImpl oldDriverConfig = rootConfig.driverConfigsById.get(newDriverConfig.id);
			if (oldDriverConfig == null) {
				newDriverConfig.activeDriver = activeDrivers.get(newDriverConfig.id);
			}
			for (DeviceConfigImpl newDeviceConfig : newDriverConfig.deviceConfigsByAddress.values()) {

				DeviceConfigImpl oldDeviceConfig = null;
				if (oldDriverConfig != null) {
					oldDeviceConfig = oldDriverConfig.deviceConfigsByAddress.get(newDeviceConfig.getIfAndDevAddress());
				}

				// add DeviceTask lists to those deviceConfigs that have
				// an interface address
				if (newDeviceConfig.interfaceAddress != null) {
					if (newDeviceConfig.driverParent.deviceTasksByInterfaceAddress
							.get(newDeviceConfig.interfaceAddress) == null) {
						if (oldDriverConfig != null) {
							List<DeviceTask> oldBusTaskList = oldDriverConfig.deviceTasksByInterfaceAddress
									.get(newDeviceConfig.interfaceAddress);
							if (oldBusTaskList != null) {
								newDeviceConfig.driverParent.deviceTasksByInterfaceAddress.put(
										newDeviceConfig.interfaceAddress, oldBusTaskList);
							}
							else {
								newDeviceConfig.driverParent.deviceTasksByInterfaceAddress.put(
										newDeviceConfig.interfaceAddress, new ArrayList<DeviceTask>());
							}
						}
						else {
							newDeviceConfig.driverParent.deviceTasksByInterfaceAddress.put(
									newDeviceConfig.interfaceAddress, new ArrayList<DeviceTask>());
						}
					}
				}

				if (oldDeviceConfig == null) {
					// Device is new
					newDeviceConfig.device = new Device(this, newDeviceConfig, currentTime, logChannels);
					if (newDeviceConfig.device.state == DeviceState.CONNECTING) {
						newDeviceConfig.device.connectRetrySignal();
					}
				}
			}
		}

		for (ChannelConfigImpl oldChannelConfig : rootConfig.channelConfigsById.values()) {
			ChannelConfigImpl newChannelConfig = newRootConfig.channelConfigsById.get(oldChannelConfig.getId());
			if (newChannelConfig == null) {
				// oldChannelConfig does not exist in the new configuration
				if (oldChannelConfig.state == ChannelState.SAMPLING) {
					removeFromSamplingCollections(oldChannelConfig.channel);
				}
				oldChannelConfig.state = ChannelState.DELETED;
				oldChannelConfig.channel.setFlag(Flag.CHANNEL_DELETED);
				// note: disabling SampleTasks and such has to be done at the
				// Device level
			}

		}

		for (LogSinkService logSink : activeLogSinks) {
			logSink.setChannelsToLog(logChannels);
		}

		newRootConfig.logChannels = logChannels;

		synchronized (configChangeListeners) {

			rootConfig = newRootConfig;
			rootConfigWithoutDefaults = configWithoutDefaults;

			for (ConfigChangeListener configChangeListener : configChangeListeners) {
				if (configChangeListener != null) {
					// TODO this should be done in a separate thread
					configChangeListener.configurationChanged();
				}
			}
		}
	}

	void addToSamplingCollections(ChannelImpl channel, Long time) {

		ChannelCollection fittingSamplingCollection = null;
		for (Action action : actions) {
			if (action.samplingCollections != null) {
				for (ChannelCollection samplingCollection : action.samplingCollections) {
					if (samplingCollection.interval == channel.config.samplingInterval
							&& samplingCollection.timeOffset == channel.config.samplingTimeOffset
							&& samplingCollection.samplingGroup.equals(channel.config.samplingGroup)
							&& samplingCollection.device == channel.config.deviceParent.device) {
						fittingSamplingCollection = samplingCollection;
						break;
					}
				}
			}
		}

		if (fittingSamplingCollection == null) {
			fittingSamplingCollection = new ChannelCollection(channel.config.samplingInterval,
					channel.config.samplingTimeOffset, channel.config.samplingGroup, channel.config.deviceParent.device);
			addSamplingCollectionToActions(fittingSamplingCollection,
					fittingSamplingCollection.calculateNextActionTime(time));
		}

		if (channel.samplingCollection != null) {
			if (channel.samplingCollection != fittingSamplingCollection) {
				removeFromSamplingCollections(channel);
			}
			else {
				return;
			}
		}
		fittingSamplingCollection.channels.add(channel);
		channel.samplingCollection = fittingSamplingCollection;
	}

	void addToLoggingCollections(ChannelImpl channel, Long time) {
		ChannelCollection fittingLoggingCollection = null;
		for (Action action : actions) {
			if (action.loggingCollections != null) {
				for (ChannelCollection loggingCollection : action.loggingCollections) {
					if (loggingCollection.interval == channel.config.loggingInterval
							&& loggingCollection.timeOffset == channel.config.loggingTimeOffset) {
						fittingLoggingCollection = loggingCollection;
						break;
					}
				}
			}
		}
		if (fittingLoggingCollection == null) {
			fittingLoggingCollection = new ChannelCollection(channel.config.loggingInterval,
					channel.config.loggingTimeOffset, null, null);
			addLoggingCollectionToActions(fittingLoggingCollection,
					fittingLoggingCollection.calculateNextActionTime(time));
		}

		if (channel.loggingCollection != null) {
			if (channel.loggingCollection != fittingLoggingCollection) {
				removeFromLoggingCollections(channel);
			}
			else {
				return;
			}
		}

		fittingLoggingCollection.channels.add(channel);
		channel.loggingCollection = fittingLoggingCollection;
	}

	void removeFromLoggingCollections(ChannelImpl channel) {
		channel.loggingCollection.channels.remove(channel);
		if (channel.loggingCollection.channels.size() == 0) {
			channel.loggingCollection.action.loggingCollections.remove(channel.loggingCollection);
		}
		channel.loggingCollection = null;
	}

	void removeFromSamplingCollections(ChannelImpl channel) {
		channel.samplingCollection.channels.remove(channel);
		if (channel.samplingCollection.channels.size() == 0) {
			channel.samplingCollection.action.samplingCollections.remove(channel.samplingCollection);
		}
		channel.samplingCollection = null;
	}

	void removeFromConnectionRetry(Device device) {
		for (Action action : actions) {
			if (action.connectionRetryDevices != null && action.connectionRetryDevices.remove(device)) {
				break;
			}
		}
	}

	protected void setDriverService(DriverService driver) {

		String driverId = driver.getId();
		logger.info("setDriverService: " + driverId);
		synchronized (newDrivers) {
			if (activeDrivers.get(driverId) != null || newDrivers.get(driverId) != null) {
				logger.error("Unable to register driver: a driver with the ID " + driverId + " is already registered.");
				return;
			}
			newDrivers.put(driverId, driver);
			interrupt();
		}
	}

	protected void unsetDriverService(DriverService driver) {

		String driverId = driver.getId();
		logger.info("Unset DriverService: " + driverId);

		// note: no synchronization needed here because this function and the
		// deactivate function are always called sequentially:
		if (dataManagerStarted) {
			driverToBeRemovedId = driverId;
			driverRemovedSignal = new CountDownLatch(1);
			interrupt();
			try {
				driverRemovedSignal.await();
			} catch (InterruptedException e) {
			}
		}
		else {
			if (activeDrivers.remove(driverId) == null) {
				newDrivers.remove(driverId);
			}
		}
	}

	protected void setLogSinkService(LogSinkService logSink) {
		String logSinkId = logSink.getId();
		logger.info("setLogSinkService: " + logSinkId);

		synchronized (newLogSinks) {
			newLogSinks.add(logSink);
			interrupt();
		}
	}

	protected void unsetLogSinkService(LogSinkService logSink) {

		String logSinkId = logSink.getId();
		logger.info("Unset LogSinkService: " + logSinkId);

		if (dataManagerStarted) {
			logSinkToBeRemoved = logSink;
			logSinkRemovedSignal = new CountDownLatch(1);
			interrupt();
			try {
				logSinkRemovedSignal.await();
			} catch (InterruptedException e) {
			}
		}
		else {
			if (activeLogSinks.remove(logSinkId) == false) {
				newLogSinks.remove(logSinkId);
			}
		}

		// logger.debug("unregistering logSink");
		//
		// logSinkRemovedSignal = new CountDownLatch(1);
		// logSinkToBeRemoved = logSink;
		// interrupt();
		//
		// try {
		// logSinkRemovedSignal.await();
		// } catch (InterruptedException e) {
		// }
		// logger.debug("removed logSink");

	}

	protected void setLogSourceService(LogSourceService logSource) {
		synchronized (newLogSources) {
			newLogSources.add(logSource);
			interrupt();
		}
	}

	protected void unsetLogSourceService(LogSourceService logSource) {
		String logSourceId = logSource.getId();
		logger.info("Unset LogSourceService: " + logSourceId);

		if (dataManagerStarted) {
			logSourceToBeRemoved = logSource;
			logSourceRemovedSignal = new CountDownLatch(1);
			interrupt();
			try {
				logSourceRemovedSignal.await();
			} catch (InterruptedException e) {
			}
		}
		else {
			if (activeLogSinks.remove(logSourceId) == false) {
				newLogSinks.remove(logSourceId);
			}
		}

		// logger.debug("unregistering logSource");
		//
		// logSourceRemovedSignal = new CountDownLatch(1);
		// logSourceToBeRemoved = logSource;
		// interrupt();
		//
		// try {
		// logSourceRemovedSignal.await();
		// } catch (InterruptedException e) {
		// }
		// logger.debug("removed logSource");
	}

	private void prepareStop() {
		// TODO tell all drivers to stop listening
		// Do I have to wait for all threads (such as SamplingTasks) to finish?
		executor.shutdown();
		dataManagerStoppedSignal.countDown();
	}

	@Override
	public Channel getChannel(String label) {
		ChannelConfigImpl channelConfig = rootConfig.channelConfigsByLabel.get(label);
		if (channelConfig == null) {
			return null;
		}
		return channelConfig.channel;
	}

	@Override
	public Channel getChannel(String label, ChannelChangeListener channelChangeListener) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public List<LogicalDevice> getLogicalDevices(String type) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public List<LogicalDevice> getLogicalDevices(String type, LogicalDeviceChangeListener logicalDeviceChangeListener) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public void newRecords(List<ChannelRecordContainer> recordContainers) {
		List<ChannelRecordContainer> recordContainersCopy = new ArrayList<ChannelRecordContainer>(
				recordContainers.size());
		for (ChannelRecordContainer container : recordContainers) {
			recordContainersCopy.add(container.copy());
		}
		synchronized (receivedRecordContainers) {
			receivedRecordContainers.add(recordContainersCopy);
		}

		interrupt();

	}

	@Override
	public void lock() {
		configLock.lock();
	}

	@Override
	public boolean tryLock() {
		return configLock.tryLock();
	}

	@Override
	public void unlock() {
		configLock.unlock();
	}

	@Override
	public RootConfig getConfig() {
		return rootConfigWithoutDefaults.clone();
	}

	@Override
	public RootConfig getConfig(ConfigChangeListener listener) {
		synchronized (configChangeListeners) {
			for (ConfigChangeListener weakChangeListener : configChangeListeners) {
				configChangeListeners.remove(weakChangeListener);
			}
			configChangeListeners.add(listener);
			return getConfig();
		}
	}

	@Override
	public void stopListeningForConfigChange(ConfigChangeListener listener) {
		synchronized (configChangeListeners) {
			for (ConfigChangeListener weakChangeListener : configChangeListeners) {
				if (weakChangeListener == listener) {
					configChangeListeners.remove(weakChangeListener);
				}
			}
		}
	}

	@Override
	public void setConfig(RootConfig config) {
		configLock.lock();
		try {
			RootConfigImpl newConfigCopy = ((RootConfigImpl) config).clone();
			setNewConfig(newConfigCopy);
		} finally {
			configLock.unlock();
		}

	}

	@Override
	public void reloadConfigFromFile() throws FileNotFoundException, ParseException {
		configLock.lock();
		try {
			RootConfigImpl newConfigCopy = RootConfigImpl.createFromFile(configFile);
			setNewConfig(newConfigCopy);
		} finally {
			configLock.unlock();
		}

	}

	private void setNewConfig(RootConfigImpl newConfigCopy) {
		synchronized (this) {
			newRootConfigWithoutDefaults = newConfigCopy;
			interrupt();
			newConfigSignal = new CountDownLatch(1);
		}
		while (true) {
			try {
				newConfigSignal.await();
				break;
			} catch (InterruptedException e) {
			}
		}

	}

	@Override
	public RootConfig getEmptyConfig() {
		return new RootConfigImpl();
	}

	@Override
	public void writeConfigToFile() throws ConfigWriteException {
		try {
			rootConfigWithoutDefaults.writeToFile(configFile);
		} catch (Exception e) {
			throw new ConfigWriteException(e);
		}
	}

	@Override
	public List<DeviceScanInformation> scanForDevices(String driverName, String scanAddress, String settings,
			int timeout) throws DriverNotAvailableException, UnsupportedOperationException, ArgumentSyntaxException,
			IOException {
		// TODO this function is probably not thread safe
		DriverService activeDriver = activeDrivers.get(driverName);
		if (activeDriver == null) {
			throw new DriverNotAvailableException();
		}
		return activeDriver.scanForDevices(null, scanAddress, settings, timeout);
	}

	@Override
	public List<ChannelScanInformation> scanForChannels(String driverName, String interfaceAddress,
			String deviceAddress, String settings, int timeout) throws DriverNotAvailableException,
			UnsupportedOperationException, IOException {
		// TODO this function is probably not thread safe
		DriverService activeDriver = activeDrivers.get(driverName);
		if (activeDriver == null) {
			throw new DriverNotAvailableException();
		}
		// TODO this has to be handled
		try {
			return activeDriver.scanForChannels(null, timeout);
		} catch (ConnectionException e) {
			throw new IOException(e);
		}
	}

	@Override
	public List<String> getIdsOfRegisteredDrivers() {
		List<String> availableDrivers = null;
		synchronized (this) {
			availableDrivers = new ArrayList<String>(activeDrivers.size());
			for (String activeDriverName : activeDrivers.keySet()) {
				availableDrivers.add(activeDriverName);
			}
		}
		return availableDrivers;
	}

	@Override
	public List<String> getIdsOfRegisteredDrivers(DriverChangeListener listener) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public void write(List<WriteValueContainer> values, int timeout) {
		HashMap<Device, List<WriteValueContainerImpl>> containersByDevice = new LinkedHashMap<Device, List<WriteValueContainerImpl>>();

		for (WriteValueContainer value : values) {
			if (value.getValue() == null) {
				((WriteValueContainerImpl) value).setFlag(Flag.CANNOT_WRITE_NULL_VALUE);
				continue;
			}
			WriteValueContainerImpl valueContainerImpl = (WriteValueContainerImpl) value;
			Device device = valueContainerImpl.channel.config.deviceParent.device;
			List<WriteValueContainerImpl> writeValueContainers = containersByDevice.get(device);
			if (writeValueContainers == null) {
				writeValueContainers = new LinkedList<WriteValueContainerImpl>();
				containersByDevice.put(device, writeValueContainers);
			}
			writeValueContainers.add(valueContainerImpl);
		}
		CountDownLatch writeTasksFinishedSignal = new CountDownLatch(containersByDevice.size());

		List<WriteTask> writeTasks = new ArrayList<WriteTask>(containersByDevice.size());
		synchronized (newWriteTasks) {
			for (Entry<Device, List<WriteValueContainerImpl>> writeValueContainers : containersByDevice.entrySet()) {
				WriteTask writeTask = new WriteTask(this, writeValueContainers.getKey(),
						writeValueContainers.getValue(), timeout / containersByDevice.size(), writeTasksFinishedSignal);
				writeTasks.add(writeTask);
				newWriteTasks.add(writeTask);
			}
		}
		interrupt();

		try {
			writeTasksFinishedSignal.await();
		} catch (InterruptedException e) {
		}

	}

	@Override
	public List<String> getAllLabels() {
		List<String> labels = new ArrayList<String>(rootConfig.channelConfigsByLabel.size());
		for (String label : rootConfig.channelConfigsByLabel.keySet()) {
			labels.add(label);
		}
		return labels;
	}

}
