/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.core.datamanager;

import java.util.List;

import org.openmuc.framework.data.Flag;
import org.openmuc.framework.dataaccess.DeviceState;
import org.openmuc.framework.driver.spi.ChannelRecordContainer;
import org.openmuc.framework.driver.spi.ConnectionException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class ReadTask extends DeviceTask {

	private final static Logger logger = LoggerFactory.getLogger(ReadTask.class);

	List<ChannelRecordContainerImpl> selectedChannels;
	private final int readTimeout;
	private boolean methodNotExcepted = false;
	private final boolean unknownDriverException = false;
	private volatile boolean disabled = false;

	boolean running = false;
	boolean startedLate = false;

	public ReadTask(DataManager dataManager, Device device, List<ChannelRecordContainerImpl> selectedChannels,
			int readTimeout) {
		this.dataManager = dataManager;
		this.device = device;
		this.selectedChannels = selectedChannels;
		this.readTimeout = readTimeout;
	}

	@Override
	@SuppressWarnings("unchecked")
	public void run() {

		try {
			device.deviceConfig.driverParent.activeDriver.read(device.connection,
					(List<ChannelRecordContainer>) ((List<?>) selectedChannels), true, null, readTimeout);
		} catch (UnsupportedOperationException e) {
			methodNotExcepted = true;
		} catch (ConnectionException e) {
			// Connection to device lost. Signal to device instance and end task without notifying DataManager
			logger.warn("Connection to device {} lost because {}. Trying to reconnect...",
					device.deviceConfig.getIfAndDevAddress(), e.getMessage());
			device.disconnectedSignal();
			return;
		} catch (Exception e) {
			logger.warn("unexpected exception thrown by read funtion of driver ", e);
			disabled = true;
		}

		// synchronized (dataManager.readTaskFinished) {
		// dataManager.readTaskFinished.add(this);
		// }
		dataManager.interrupt();
	}

	public void timeout() {
		if (!disabled) {
			disabled = true;
			if (startedLate) {
				for (ChannelRecordContainerImpl driverChannel : selectedChannels) {
					driverChannel.channel.setFlag(Flag.STARTED_LATE_AND_TIMED_OUT);
				}
			}
			else if (running) {
				for (ChannelRecordContainerImpl driverChannel : selectedChannels) {
					driverChannel.channel.setFlag(Flag.TIMEOUT);
				}
			}
			else {
				for (ChannelRecordContainerImpl driverChannel : selectedChannels) {
					driverChannel.channel.setFlag(Flag.DEVICE_OR_INTERFACE_BUSY);
				}
				// device.removeTask(this);
			}

		}
	}

	public void storeValues() {
		if (disabled) {
			return;
		}
		if (methodNotExcepted) {
			for (ChannelRecordContainerImpl driverChannel : selectedChannels) {
				driverChannel.channel.setFlag(Flag.ACCESS_METHOD_NOT_SUPPORTED);
			}
		}
		else if (unknownDriverException) {
			for (ChannelRecordContainerImpl driverChannel : selectedChannels) {
				driverChannel.channel.setFlag(Flag.DRIVER_THREW_UNKNOWN_EXCEPTION);
			}
		}
		else {
			for (ChannelRecordContainerImpl driverChannel : selectedChannels) {
				driverChannel.channel.setNewSample(driverChannel.record);
			}
		}
		disabled = true;

	}

	@Override
	public DeviceTaskType getType() {
		return DeviceTaskType.READ;
	}

	@Override
	public void setDeviceState() {
		device.state = DeviceState.READING;
	}

}
