/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.datalogger.ascii;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

import org.openmuc.framework.data.Flag;
import org.openmuc.framework.data.TypeConversionException;
import org.openmuc.framework.data.ValueType;
import org.openmuc.framework.datalogger.spi.LogChannel;
import org.openmuc.framework.datalogger.spi.LogRecordContainer;
import org.openmuc.framework.datalogger.spi.LogSinkService;
import org.osgi.service.component.ComponentContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class AsciiLogger implements LogSinkService {

	private static final Logger logger = LoggerFactory.getLogger(AsciiLogger.class);

	private static final String EXTENSION = ".dat";
	private static final String DEFAULT_DIR = "asciidata/";

	private static final String DIR = System.getProperty(AsciiLogger.class.getPackage().getName().toLowerCase()
			+ ".directory");

	private long lastTimestamp = 0;
	private int minLoggingInterval;
	private List<String> idList;
	private List<String> labelList;
	private List<ValueType> typeList;

	private FileOutHandler fileOut;

	@Override
	public String getId() {
		return "asciilogger";
	}

	protected void activate(ComponentContext context) {
		logger.info("Activating AsciiLogger");

		String rootFolder = AsciiLogger.DIR;
		if (rootFolder == null) {
			rootFolder = DEFAULT_DIR;
			logger.info("System property " + AsciiLogger.class.getPackage().getName().toLowerCase()
					+ ".directory not set. Storing to " + AsciiLogger.DEFAULT_DIR);
		}

		fileOut = new FileOutHandler(rootFolder, EXTENSION);
	}

	protected void deactivate(ComponentContext context) {
		logger.info("Deactivating AsciiLogger");

		fileOut.close();
	}

	@Override
	public void setChannelsToLog(List<LogChannel> channels) {
		minLoggingInterval = Integer.MAX_VALUE;

		idList = new ArrayList<String>(channels.size());
		typeList = new ArrayList<ValueType>(channels.size());
		List<String> newLabelList = new ArrayList<String>(channels.size());

		for (LogChannel channel : channels) {
			if (isLoggable(channel.getValueType())) {
				minLoggingInterval = Math.min(minLoggingInterval, channel.getLoggingInterval());
				String label = null;

				if (channel.getLabel() != null && channel.getLabel().isEmpty() == false) {
					label = channel.getLabel().replace(" ", "_").replace("ä", "ae").replace("ö", "oe")
							.replace("ü", "ue").replace("Ä", "AE").replace("Ö", "OE").replace("Ü", "UE")
							.replace("ß", "ss");
					if (channel.getUnit() != null && channel.getUnit().isEmpty() == false) {
						label += "_[" + channel.getUnit() + "]";
					}
				}
				else {
					label = channel.getId();
				}

				idList.add(channel.getId());
				typeList.add(channel.getValueType());
				newLabelList.add(label);
			}
		}

		if (newLabelList.equals(labelList) == false) {
			labelList = newLabelList;
			fileOut.writeHeader(labelList);
		}
	}

	@Override
	public void log(List<LogRecordContainer> containers, long timestamp) {
		Calendar logTime = Calendar.getInstance();
		logTime.setTimeInMillis(timestamp);

		List<String> values = createRecordSet(logTime);

		if (lastTimestamp + minLoggingInterval <= timestamp) {
			for (LogRecordContainer container : containers) {
				if (container.getRecord().getFlag() == Flag.VALID) {
					int index = idList.indexOf(container.getChannelId());
					try {
						ValueType type = typeList.get(index);
						if (isFloatingPoint(type)) {
							values.set(index + 2, String.format("%.2f", container.getRecord().getValue().asDouble()));
						}
						else if (isIntegral(type)) {
							values.set(index + 2, String.format("%d", container.getRecord().getValue().asLong()));
						}
					} catch (TypeConversionException ex) {
						logger.debug("Error converting channel " + container.getChannelId());
					}
				}
			}

			fileOut.writeRecords(labelList, values, logTime);
			lastTimestamp = timestamp;
		}
	}

	private List<String> createRecordSet(Calendar time) {
		String timeValue = String.format("%02d%02d%02d", time.get(Calendar.HOUR_OF_DAY), time.get(Calendar.MINUTE),
				time.get(Calendar.SECOND));

		List<String> result = new ArrayList<String>(idList.size() + 2);
		result.add(timeValue);
		result.add(Long.toString(time.getTimeInMillis() / 1000));

		for (int i = 0; i < idList.size(); i++) {
			result.add("-99.9");
		}

		return result;
	}

	private boolean isLoggable(ValueType type) {
		return isFloatingPoint(type) || isIntegral(type);
	}

	private boolean isIntegral(ValueType type) {
		return type == ValueType.BYTE || type == ValueType.SHORT || type == ValueType.INTEGER || type == ValueType.LONG;
	}

	private boolean isFloatingPoint(ValueType type) {
		return type == ValueType.FLOAT || type == ValueType.DOUBLE;
	}
}
