/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.datalogger.ascii;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Calendar;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class FileOutHandler {

	private static final Logger logger = LoggerFactory.getLogger(FileOutHandler.class);

	private String filename;
	private PrintStream out;

	private final Calendar fileValidUntil = Calendar.getInstance();

	private final String path;
	private final String extension;

	public FileOutHandler(String path, String extension) {
		this.path = path;
		this.extension = extension;

		// Set the time until the current file is valid to 0:00:00.0 of today. This way the first call of fileOutdated()
		// always returns true
		fileValidUntil.set(Calendar.HOUR_OF_DAY, 0);
		fileValidUntil.set(Calendar.MINUTE, 0);
		fileValidUntil.set(Calendar.SECOND, 0);
		fileValidUntil.set(Calendar.MILLISECOND, 0);
	}

	public void writeHeader(List<String> labels) {
		if (out != null) {
			out.print("#TIME TIME_DEC ");
			for (String label : labels) {
				out.print(label + " ");
			}
			out.println();
			out.println("#");

			out.flush();
		}
	}

	public void writeRecords(List<String> labels, List<String> records, Calendar logTime) {
		if (fileOutdated(logTime)) {
			createFile();
			writeHeader(labels);
		}

		if (out == null) {
			return;
		}

		for (String record : records) {
			out.print(record + " ");
		}
		out.println();
		out.flush();
	}

	public void close() {
		if (out != null) {
			out.close();
		}
	}

	private boolean fileOutdated(Calendar logTime) {
		if (logTime.before(fileValidUntil)) {
			return false;
		}

		return true;
	}

	private void createFile() {
		Calendar date = Calendar.getInstance();

		filename = String.format("%02d%02d%02d%s", date.get(Calendar.YEAR) % 100, date.get(Calendar.MONTH) + 1,
				date.get(Calendar.DAY_OF_MONTH), extension);

		// Check if an old file is still open
		if (out != null) {
			out.close();
			out = null;
		}

		try {
			File folder = new File(path);
			if (!folder.exists()) {
				folder.mkdir();
			}
			if (!folder.isDirectory()) {
				throw new IOException(path + " is no directory");
			}

			File file = new File(path, filename);

			// file.createNewFile() checks if the file already exists, if it does, nothing happens at this point
			file.createNewFile();

			// opens the file, appending new data
			out = new PrintStream(new FileOutputStream(file, true), false, "US-ASCII");

			fileValidUntil.add(Calendar.DAY_OF_MONTH, 1);
		} catch (IOException e) {
			filename = null;
			out = null;
			logger.error("Cannot create file " + filename + ". Reason: " + e.getMessage());
		}
	}
}
