/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.driver.iec61850;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.openmuc.framework.config.ArgumentSyntaxException;
import org.openmuc.framework.config.ChannelScanInformation;
import org.openmuc.framework.config.DeviceScanInformation;
import org.openmuc.framework.data.BooleanValue;
import org.openmuc.framework.data.ByteStringValue;
import org.openmuc.framework.data.ByteValue;
import org.openmuc.framework.data.DoubleValue;
import org.openmuc.framework.data.Flag;
import org.openmuc.framework.data.FloatValue;
import org.openmuc.framework.data.IntValue;
import org.openmuc.framework.data.LongValue;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.data.ShortValue;
import org.openmuc.framework.driver.spi.ChannelRecordContainer;
import org.openmuc.framework.driver.spi.ChannelValueContainer;
import org.openmuc.framework.driver.spi.ConnectionException;
import org.openmuc.framework.driver.spi.DeviceConnection;
import org.openmuc.framework.driver.spi.DriverService;
import org.openmuc.framework.driver.spi.RecordsReceivedListener;
import org.openmuc.openiec61850.BasicDataAttribute;
import org.openmuc.openiec61850.BdaBoolean;
import org.openmuc.openiec61850.BdaFloat32;
import org.openmuc.openiec61850.BdaFloat64;
import org.openmuc.openiec61850.BdaTimestamp;
import org.openmuc.openiec61850.ClientAssociation;
import org.openmuc.openiec61850.ClientSap;
import org.openmuc.openiec61850.FcModelNode;
import org.openmuc.openiec61850.ModelNode;
import org.openmuc.openiec61850.ServerNode;
import org.openmuc.openiec61850.ServiceError;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class Iec61850Driver implements DriverService {

	private static Logger logger = LoggerFactory.getLogger(Iec61850Driver.class);

	private final static String deviceAddressSyntax = "url:port[:authenticationParameter]";

	@Override
	public String getId() {
		return "iec61850";
	}

	@Override
	public String getDescription() {
		return "This driver can be used to access devices using IEC 61850 (MMS)";
	}

	@Override
	public List<DeviceScanInformation> scanForDevices(String interfaceAddress, String scanFilter, String settings,
			int timeout) throws UnsupportedOperationException, ArgumentSyntaxException, IOException {
		throw new UnsupportedOperationException();
	}

	@Override
	public List<ChannelScanInformation> scanForChannels(DeviceConnection connection, int timeout)
			throws UnsupportedOperationException, ConnectionException {
		throw new UnsupportedOperationException();
	}

	@Override
	public Object connect(String interfaceAddress, String deviceAddress, String settings, int timeout)
			throws ConnectionException {

		String[] deviceAddresses = deviceAddress.split(":");

		if (deviceAddresses.length < 2 || deviceAddresses.length > 3) {
			throw new ConnectionException("invalid device address syntax, correct syntax: " + deviceAddressSyntax);
		}

		String remoteHost = deviceAddresses[0];
		InetAddress address;
		try {
			address = InetAddress.getByName(remoteHost);
		} catch (UnknownHostException e) {
			throw new ConnectionException("Unknown host: " + remoteHost, e);
		}

		int remotePort;
		try {
			remotePort = Integer.parseInt(deviceAddresses[1]);
		} catch (NumberFormatException e) {
			throw new ConnectionException("invalid port specified", e);
		}

		ClientSap clientSap = new ClientSap();

		if (settings != null && !settings.isEmpty()) {
			String[] splitSettings = settings.split(";");
			for (String setting : splitSettings) {
				String[] splitSetting = setting.split("=");
				if (splitSetting.length < 1 || splitSetting.length > 2) {
					throw new ConnectionException("settings syntax incorrect");
				}
				if (splitSetting[0].equalsIgnoreCase("TSelLocal")) {
					if (splitSetting.length == 1) {
						clientSap.setTSelLocal(null);
					}
					else {
						byte[] tSelLocal = new byte[splitSetting[1].length()];
						for (int i = 0; i < splitSetting[1].length(); i++) {
							tSelLocal[i] = (byte) splitSetting[1].charAt(i);
						}
						clientSap.setTSelLocal(tSelLocal);
					}
				}
				else if (splitSetting[0].equalsIgnoreCase("TSelRemote")) {
					if (splitSetting.length == 1) {
						clientSap.setTSelRemote(null);
					}
					else {
						byte[] tSelRemote = new byte[splitSetting[1].length()];
						for (int i = 0; i < splitSetting[1].length(); i++) {
							tSelRemote[i] = (byte) splitSetting[1].charAt(i);
						}
						clientSap.setTSelRemote(tSelRemote);
					}
				}
			}
		}

		ClientAssociation clientAssociation;
		try {
			clientAssociation = clientSap.associate(address, remotePort, null);
		} catch (IOException e) {
			throw new ConnectionException(e);
		}

		ServerNode serverNode;
		try {
			serverNode = clientAssociation.retrieveModel();
		} catch (ServiceError e) {
			clientAssociation.close();
			throw new ConnectionException("Service error retrieving server model" + e.getMessage(), e);
		} catch (IOException e) {
			clientAssociation.close();
			throw new ConnectionException("IOException retrieving server model: " + e.getMessage(), e);
		}

		return new Iec61850Connection(clientAssociation, serverNode);
	}

	@Override
	public void disconnect(DeviceConnection connection) {
		((Iec61850Connection) connection.getConnectionHandle()).getClientAssociation().close();
	}

	@Override
	public Object read(DeviceConnection connection, List<ChannelRecordContainer> containers,
			Object containerListHandle, String samplingGroup, int timeout) throws UnsupportedOperationException,
			ConnectionException {

		Iec61850Connection connectionHandle = (Iec61850Connection) connection.getConnectionHandle();
		ServerNode serverNode = connectionHandle.getServerNode();

		for (ChannelRecordContainer container : containers) {

			if (container.getChannelHandle() == null) {
				ModelNode modelNode;
				if (container.getChannelAddress().endsWith(":bytestring")) {
					modelNode = serverNode.findSubNode(container.getChannelAddress().substring(0,
							container.getChannelAddress().length() - 11));
				}
				else {
					modelNode = serverNode.findSubNode(container.getChannelAddress());
				}

				if (modelNode == null) {
					logger.debug(
							"Error reading channel: no FCDO/DA with object reference {} was not found in the server model.",
							container.getChannelAddress());
					container.setRecord(new Record(null, null, Flag.DRIVER_ERROR_CHANNEL_WITH_THIS_ADDRESS_NOT_FOUND));
					continue;
				}

				BasicDataAttribute bda;
				try {
					bda = (BasicDataAttribute) modelNode;
				} catch (ClassCastException e) {
					logger.debug(
							"Error reading channel: ModelNode with object reference {} was found in the server model but is not a Basic Data Attribute.",
							container.getChannelAddress());
					container.setRecord(new Record(null, null, Flag.DRIVER_ERROR_CHANNEL_WITH_THIS_ADDRESS_NOT_FOUND));
					continue;
				}

				container.setChannelHandle(bda);

			}
		}

		if (!samplingGroup.isEmpty()) {

			FcModelNode fcModelNode;
			if (containerListHandle != null) {
				fcModelNode = (FcModelNode) containerListHandle;
			}

			else {

				ModelNode modelNode = serverNode.findSubNode(samplingGroup);

				if (modelNode == null) {
					logger.debug(
							"Error reading sampling group: no FCDO/DA or DataSet with object reference {} was not found in the server model.",
							samplingGroup);
					for (ChannelRecordContainer container : containers) {
						container.setRecord(new Record(null, null, Flag.DRIVER_ERROR_SAMPLING_GROUP_NOT_FOUND));
					}
					return null;
				}

				try {
					fcModelNode = (FcModelNode) modelNode;
				} catch (ClassCastException e) {
					logger.debug(
							"Error reading channel: ModelNode with sampling group reference {} was found in the server model but is not a FcModelNode.",
							samplingGroup);
					for (ChannelRecordContainer container : containers) {
						container.setRecord(new Record(null, null, Flag.DRIVER_ERROR_SAMPLING_GROUP_NOT_FOUND));
					}
					return null;
				}

			}

			for (ChannelRecordContainer container : containers) {
				if (container.getChannelHandle() != null) {
					BasicDataAttribute bda = (BasicDataAttribute) container.getChannelHandle();
					bda.setValue(null);
				}

			}

			try {
				connectionHandle.getClientAssociation().getDataValues(fcModelNode);
			} catch (ServiceError e) {
				if (e.getErrorCode() == ServiceError.CONNECTION_LOST) {
					throw new ConnectionException(e);
				}
				logger.debug("Error reading sampling group: service error calling getDataValues on {}: {}",
						samplingGroup, e);
				for (ChannelRecordContainer container : containers) {
					container.setRecord(new Record(null, null, Flag.DRIVER_ERROR_SAMPLING_GROUP_NOT_ACCESSIBLE));
				}
				return fcModelNode;
			} catch (IOException e) {
				throw new ConnectionException();
			}

			long receiveTime = System.currentTimeMillis();

			for (ChannelRecordContainer container : containers) {
				if (container.getChannelHandle() != null) {
					BasicDataAttribute bda = (BasicDataAttribute) container.getChannelHandle();
					setRecord(container, bda, receiveTime);
				}
				else {
					container.setRecord(new Record(null, null, Flag.DRIVER_ERROR_CHANNEL_NOT_PART_OF_SAMPLING_GROUP));
				}
			}

			return fcModelNode;

		}
		else {

			for (ChannelRecordContainer container : containers) {

				if (container.getChannelHandle() == null) {
					continue;
				}
				BasicDataAttribute bda = (BasicDataAttribute) container.getChannelHandle();
				try {
					connectionHandle.getClientAssociation().getDataValues(bda);
				} catch (ServiceError e) {
					logger.debug("Error reading channel: service error calling getDataValues on {}: {}",
							container.getChannelAddress(), e);
					container.setRecord(new Record(null, null, Flag.DRIVER_ERROR_CHANNEL_NOT_ACCESSIBLE));
					continue;
				} catch (IOException e) {
					throw new ConnectionException(e);
				}

				long receiveTime = System.currentTimeMillis();

				setRecord(container, bda, receiveTime);

			}
			return null;
		}

	}

	private void setRecord(ChannelRecordContainer container, BasicDataAttribute bda, long receiveTime) {
		switch (bda.getBasicType()) {
		case BIT_STRING:
		case ENTRY_TIME:
		case OCTET_STRING:
		case VISIBLE_STRING:
		case UNICODE_STRING:
			container.setRecord(new Record(new ByteStringValue((byte[]) bda.getValue()), receiveTime));
			break;
		case TIMESTAMP:
			if (container.getChannelAddress().endsWith(":bytestring")) {
				container.setRecord(new Record(new ByteStringValue((byte[]) bda.getValue()), receiveTime));
			}
			else {
				container.setRecord(new Record(new LongValue(((BdaTimestamp) bda).getDate().getTime()), receiveTime));
			}
			break;
		case BOOLEAN:
			container.setRecord(new Record(new BooleanValue(((BdaBoolean) bda).getValue()), receiveTime));
			break;
		case FLOAT32:
			container.setRecord(new Record(new FloatValue(((BdaFloat32) bda).getFloat()), receiveTime));
			break;
		case FLOAT64:
			container.setRecord(new Record(new DoubleValue(((BdaFloat64) bda).getDouble()), receiveTime));
			break;
		case INT8:
			container.setRecord(new Record(new ByteValue((Byte) bda.getValue()), receiveTime));
			break;
		case INT8U:
		case INT16:
			container.setRecord(new Record(new ShortValue((Short) bda.getValue()), receiveTime));
			break;
		case INT16U:
		case INT32:
			container.setRecord(new Record(new IntValue((Integer) bda.getValue()), receiveTime));
			break;
		case INT32U:
		case INT64:
			container.setRecord(new Record(new LongValue((Long) bda.getValue()), receiveTime));
			break;
		default:
			throw new IllegalStateException("unknown BasicType received: " + bda.getBasicType());
		}
	}

	@Override
	public void startListening(DeviceConnection connection, List<ChannelRecordContainer> containers,
			RecordsReceivedListener listener) throws UnsupportedOperationException {
		throw new UnsupportedOperationException();
	}

	@Override
	public Object write(DeviceConnection connection, List<ChannelValueContainer> containers,
			Object containerListHandle, int timeout) throws UnsupportedOperationException, ConnectionException {

		Iec61850Connection connectionHandle = (Iec61850Connection) connection.getConnectionHandle();
		ServerNode serverNode = connectionHandle.getServerNode();

		List<BasicDataAttribute> remainingContainers = new ArrayList<BasicDataAttribute>(containers.size());

		for (ChannelValueContainer container : containers) {

			if (container.getChannelHandle() != null) {
				remainingContainers.add((BasicDataAttribute) container.getChannelHandle());
				setBda(container, (BasicDataAttribute) container.getChannelHandle());
			}
			else {

				ModelNode modelNode;
				if (container.getChannelAddress().endsWith(":bytestring")) {
					modelNode = serverNode.findSubNode(container.getChannelAddress().substring(0,
							container.getChannelAddress().length() - 11));
				}
				else {
					modelNode = serverNode.findSubNode(container.getChannelAddress());
				}

				if (modelNode == null) {
					logger.debug(
							"Error reading channel: no FCDO/DA with object reference {} was not found in the server model.",
							container.getChannelAddress());
					container.setFlag(Flag.DRIVER_ERROR_CHANNEL_WITH_THIS_ADDRESS_NOT_FOUND);
					continue;
				}

				BasicDataAttribute bda;
				try {
					bda = (BasicDataAttribute) modelNode;
				} catch (ClassCastException e) {
					logger.debug(
							"Error reading channel: ModelNode with object reference {} was found in the server model but is not a Basic Data Attribute.",
							container.getChannelAddress());
					container.setFlag(Flag.DRIVER_ERROR_CHANNEL_WITH_THIS_ADDRESS_NOT_FOUND);
					continue;
				}

				container.setChannelHandle(bda);
				remainingContainers.add(bda);
				setBda(container, bda);

			}
		}

		// TODO
		// first check all datasets if the are some that contain only requested channels
		// then check all remaining model nodes

		List<FcModelNode> fcNodesToBeRequested = new ArrayList<FcModelNode>();

		while (remainingContainers.size() > 0) {
			fillRequestedNodes(fcNodesToBeRequested, remainingContainers, serverNode);
		}

		for (FcModelNode fcModelNode : fcNodesToBeRequested) {
			try {
				connectionHandle.getClientAssociation().setDataValues(fcModelNode);
			} catch (ServiceError e) {
				logger.debug("Error writing to channel: service error calling setDataValues on {}: {}",
						fcModelNode.getReference(), e);
				for (BasicDataAttribute bda : fcModelNode.getBasicDataAttributes()) {
					for (ChannelValueContainer valueContainer : containers) {
						if (valueContainer.getChannelHandle() == bda) {
							valueContainer.setFlag(Flag.DRIVER_ERROR_CHANNEL_NOT_ACCESSIBLE);
						}
					}
				}
			} catch (IOException e) {
				throw new ConnectionException(e);
			}
			for (BasicDataAttribute bda : fcModelNode.getBasicDataAttributes()) {
				for (ChannelValueContainer valueContainer : containers) {
					if (valueContainer.getChannelHandle() == bda) {
						valueContainer.setFlag(Flag.VALID);
					}
				}
			}
		}

		return null;

	}

	void fillRequestedNodes(List<FcModelNode> fcNodesToBeRequested, List<BasicDataAttribute> remainingBdas,
			ServerNode serverNode) {

		BasicDataAttribute currentBda = remainingBdas.get(0);

		if (!checkParent(currentBda, fcNodesToBeRequested, remainingBdas, serverNode)) {
			remainingBdas.remove(currentBda);
			fcNodesToBeRequested.add(currentBda);
		}

	}

	boolean checkParent(ModelNode modelNode, List<FcModelNode> fcNodesToBeRequested,
			List<BasicDataAttribute> remainingBdas, ServerNode serverNode) {

		if (!(modelNode instanceof FcModelNode)) {
			return false;
		}

		ModelNode parentNode = serverNode;
		for (int i = 0; i < modelNode.getReference().size() - 1; i++) {
			parentNode = parentNode.getChild(modelNode.getReference().get(i));
		}

		List<BasicDataAttribute> basicDataAttributes = parentNode.getBasicDataAttributes();
		for (BasicDataAttribute bda : basicDataAttributes) {
			if (!remainingBdas.contains(bda)) {
				return false;
			}
		}

		if (!checkParent(parentNode, fcNodesToBeRequested, remainingBdas, serverNode)) {
			for (BasicDataAttribute bda : basicDataAttributes) {
				remainingBdas.remove(bda);
			}
			fcNodesToBeRequested.add((FcModelNode) parentNode);
		}

		return true;
	}

	private void setBda(ChannelValueContainer container, BasicDataAttribute bda) {
		switch (bda.getBasicType()) {
		case BIT_STRING:
		case ENTRY_TIME:
		case OCTET_STRING:
		case VISIBLE_STRING:
		case UNICODE_STRING:
			bda.setValue(container.getValue().asByteArray());
			break;
		case TIMESTAMP:
			if (container.getChannelAddress().endsWith(":bytestring")) {
				bda.setValue(container.getValue().asByteArray());
			}
			else {
				((BdaTimestamp) bda).setDate(new Date(container.getValue().asLong()));
			}
			break;
		case BOOLEAN:
			bda.setValue(container.getValue().asBoolean());
			break;
		case FLOAT32:
			bda.setValue(container.getValue().asFloat());
			break;
		case FLOAT64:
			bda.setValue(container.getValue().asDouble());
			break;
		case INT8:
			bda.setValue(container.getValue().asByte());
			break;
		case INT8U:
		case INT16:
			bda.setValue(container.getValue().asShort());
			break;
		case INT16U:
		case INT32:
			bda.setValue(container.getValue().asInt());
			break;
		case INT32U:
		case INT64:
			bda.setValue(container.getValue().asLong());
			break;
		default:
			throw new IllegalStateException("unknown BasicType received: " + bda.getBasicType());
		}
	}

}
