/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.driver.knx;

import java.net.InetSocketAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;

import org.openmuc.framework.config.ArgumentSyntaxException;
import org.openmuc.framework.config.ChannelScanInformation;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.driver.spi.ChannelRecordContainer;
import org.openmuc.framework.driver.spi.ConnectionException;
import org.openmuc.framework.driver.spi.RecordsReceivedListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import tuwien.auto.calimero.GroupAddress;
import tuwien.auto.calimero.exception.KNXException;
import tuwien.auto.calimero.link.KNXLinkClosedException;
import tuwien.auto.calimero.link.KNXNetworkLink;
import tuwien.auto.calimero.link.KNXNetworkLinkIP;
import tuwien.auto.calimero.link.medium.KNXMediumSettings;
import tuwien.auto.calimero.link.medium.TPSettings;
import tuwien.auto.calimero.process.ProcessCommunicator;
import tuwien.auto.calimero.process.ProcessCommunicatorImpl;

public class KnxConnection {

	private static Logger logger = LoggerFactory.getLogger(KnxConnection.class);
	private static int DEFAULT_PORT = 3671;
	private static int DEFAULT_TIMEOUT = 2;

	private KNXNetworkLink knxNetworkLink;
	private ProcessCommunicator processCommunicator;
	private KnxProcessListener processListener;
	private int responseTimeout;

	public KnxConnection(String interfaceAddress, String deviceAddress, String settings, int timeout)
			throws ArgumentSyntaxException, ConnectionException {
		// TODO connect over ip or serial?

		connectNetIP(interfaceAddress, deviceAddress);
		try {
			processCommunicator = new ProcessCommunicatorImpl(knxNetworkLink);
			processListener = new KnxProcessListener();
			processCommunicator.addProcessListener(processListener);
			setResponseTimeout(timeout);
		} catch (KNXLinkClosedException e) {
			e.printStackTrace();
			throw new ConnectionException(e);
		}
	}

	private void connectNetIP(String localEP, String remoteEP) throws ArgumentSyntaxException, ConnectionException {

		try {
			URI localUri = new URI("knx://" + localEP);
			String localIP = localUri.getHost();
			int localPort = localUri.getPort() < 0 ? DEFAULT_PORT : localUri.getPort();

			URI remoteUri = new URI("knx://" + remoteEP);
			String remoteIP = remoteUri.getHost();
			int remotePort = remoteUri.getPort() < 0 ? DEFAULT_PORT : remoteUri.getPort();

			int serviceMode = KNXNetworkLinkIP.TUNNELING;
			InetSocketAddress localSocket = new InetSocketAddress(localIP, localPort);
			InetSocketAddress remoteSocket = new InetSocketAddress(remoteIP, remotePort);
			boolean useNAT = true;
			KNXMediumSettings settings = new TPSettings(true);

			knxNetworkLink = new KNXNetworkLinkIP(serviceMode, localSocket, remoteSocket, useNAT, settings);
		} catch (URISyntaxException e) {
			logger.warn("The address has the wrong format");
			throw new ArgumentSyntaxException();
		} catch (KNXException e) {
			logger.warn("Connection failed: " + e.getMessage());
			throw new ConnectionException(e);
		} catch (InterruptedException e) {
			e.printStackTrace();
			throw new ConnectionException(e);
		}

	}

	public void disconnect() {
		processCommunicator.detach();
		knxNetworkLink.close();
	}

	public List<ChannelScanInformation> scanForChannels(int timeout) {
		List<ChannelScanInformation> informations = new LinkedList<ChannelScanInformation>();
		Hashtable<GroupAddress, byte[]> values = processListener.getCachedValues();
		Enumeration<GroupAddress> keys = values.keys();

		while (keys.hasMoreElements()) {
			GroupAddress groupAddress = keys.nextElement();
			byte[] asdu = values.get(groupAddress);
			String channelAddress = groupAddress.toString() + ":1.001";
			String value = "";
			for (byte b : asdu) {
				value += String.format("%x", b);
			}
			String description = "Datapoint length: " + asdu.length + "; Last datapoint value: 0x" + value;
			informations.add(new ChannelScanInformation(channelAddress, description, null, null));
		}
		return informations;
	}

	public Record read(KnxGroupDP groupDP, int timeout) throws ConnectionException, KNXException {
		Record record = null;
		if (responseTimeout != timeout) {
			setResponseTimeout(timeout);
		}
		try {
			groupDP.getKnxValue().setDPTValue(processCommunicator.read(groupDP));

			record = new Record(groupDP.getKnxValue().getOpenMucValue(), System.currentTimeMillis());
		} catch (InterruptedException e) {
			e.printStackTrace();
			throw new ConnectionException("Read failed for group address " + groupDP.getMainAddress());
		}

		return record;
	}

	public boolean write(KnxGroupDP groupDP, int timeout) throws KNXException {
		if (responseTimeout != timeout) {
			setResponseTimeout(timeout);
		}
		try {
			processCommunicator.write(groupDP, groupDP.getKnxValue().getDPTValue());
			return true;
		} catch (KNXException e) {
			logger.warn("write failed");
			return false;
		}
	}

	public void startListening(List<ChannelRecordContainer> containers, RecordsReceivedListener listener) {
		processListener.registerOpenMucListener(containers, listener);
	}

	private void setResponseTimeout(int timeout) {
		responseTimeout = timeout;
		int timeoutSec = (timeout / 1000);
		if (timeoutSec > 0) {
			processCommunicator.setResponseTimeout(timeoutSec);
		}
		else {
			processCommunicator.setResponseTimeout(DEFAULT_TIMEOUT);
		}
	}

}
