/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.driver.knx;

import java.io.IOException;
import java.util.List;

import org.openmuc.framework.config.ArgumentSyntaxException;
import org.openmuc.framework.config.ChannelScanInformation;
import org.openmuc.framework.config.DeviceScanInformation;
import org.openmuc.framework.data.Flag;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.driver.spi.ChannelRecordContainer;
import org.openmuc.framework.driver.spi.ChannelValueContainer;
import org.openmuc.framework.driver.spi.ConnectionException;
import org.openmuc.framework.driver.spi.DeviceConnection;
import org.openmuc.framework.driver.spi.DriverService;
import org.openmuc.framework.driver.spi.RecordsReceivedListener;
import org.osgi.service.component.ComponentContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import tuwien.auto.calimero.GroupAddress;
import tuwien.auto.calimero.exception.KNXException;
import tuwien.auto.calimero.exception.KNXTimeoutException;
import tuwien.auto.calimero.log.LogManager;

public class KnxDriver implements DriverService {

	private static Logger logger = LoggerFactory.getLogger(KnxDriver.class);

	protected void activate(ComponentContext context) {
		if (logger.isDebugEnabled()) {
			LogManager.getManager().addWriter("", new KnxLogWriter()); // Add calimero logger
		}
	}

	protected void deactivate(ComponentContext context) {
		logger.debug("Deactivated KNX Driver");
	}

	@Override
	public String getId() {
		return "knx";
	}

	@Override
	public String getDescription() {
		return "Driver to read and write KNX groupaddresses";
	}

	@Override
	public List<DeviceScanInformation> scanForDevices(String interfaceAddress, String scanFilter, String settings,
			int timeout) throws UnsupportedOperationException, ArgumentSyntaxException, IOException {
		boolean natAware = false;
		boolean mcastResponse = false;
		if (settings != null) {
			logger.debug("Applying settings: " + settings);
			natAware = settings.contains("nat");
			mcastResponse = settings.contains("mcast");
		}
		KnxIpDiscover discover = new KnxIpDiscover(interfaceAddress, natAware, mcastResponse);
		discover.startSearch(timeout);
		return discover.getResponses();
	}

	@Override
	public List<ChannelScanInformation> scanForChannels(DeviceConnection connection, int timeout)
			throws UnsupportedOperationException, ConnectionException {
		return ((KnxConnection) connection.getConnectionHandle()).scanForChannels(timeout);
	}

	@Override
	public Object connect(String interfaceAddress, String deviceAddress, String settings, int timeout)
			throws ArgumentSyntaxException, ConnectionException {
		logger.debug("Connecting from " + interfaceAddress + " to " + deviceAddress + " " + settings + " " + timeout);
		return new KnxConnection(interfaceAddress, deviceAddress, settings, timeout);
	}

	@Override
	public void disconnect(DeviceConnection connection) {
		logger.debug("Disconnecting from " + connection.getInterfaceAddress() + " " + connection.getDeviceAddress());
		((KnxConnection) connection.getConnectionHandle()).disconnect();
	}

	@Override
	public Object read(DeviceConnection connection, List<ChannelRecordContainer> containers,
			Object containerListHandle, String samplingGroup, int timeout) throws UnsupportedOperationException,
			ConnectionException {

		KnxConnection knxConnection = (KnxConnection) connection.getConnectionHandle();

		for (ChannelRecordContainer container : containers) {
			try {

				KnxGroupDP groupDP = null;
				if (container.getChannelHandle() == null) {
					groupDP = createKnxGroupDP(container.getChannelAddress());
					logger.debug("New datapoint: " + groupDP);
					container.setChannelHandle(groupDP);
				}
				else {
					groupDP = (KnxGroupDP) container.getChannelHandle();
				}

				Record record = knxConnection.read(groupDP, timeout);
				container.setRecord(record);
			} catch (KNXTimeoutException e1) {
				logger.debug(e1.getMessage());
				container.setRecord(new Record(null, System.currentTimeMillis(), Flag.TIMEOUT));
			} catch (KNXException e) {
				e.printStackTrace();
			}

		}

		return null;
	}

	@Override
	public void startListening(DeviceConnection connection, List<ChannelRecordContainer> containers,
			RecordsReceivedListener listener) throws UnsupportedOperationException, ConnectionException {
		for (ChannelRecordContainer container : containers) {
			if (container.getChannelHandle() == null) {
				try {
					container.setChannelHandle(createKnxGroupDP(container.getChannelAddress()));
				} catch (KNXException e) {
					e.printStackTrace();
				}
			}
		}
		logger.info("Start listening for " + containers.size() + " channels");
		((KnxConnection) connection.getConnectionHandle()).startListening(containers, listener);
	}

	@Override
	public Object write(DeviceConnection connection, List<ChannelValueContainer> containers,
			Object containerListHandle, int timeout) throws UnsupportedOperationException, ConnectionException {

		KnxConnection knxConnection = (KnxConnection) connection.getConnectionHandle();

		for (ChannelValueContainer container : containers) {
			try {
				KnxGroupDP groupDP = null;
				if (container.getChannelHandle() == null) {
					groupDP = createKnxGroupDP(container.getChannelAddress());
					logger.debug("New datapoint: " + groupDP);
					container.setChannelHandle(groupDP);
				}
				else {
					groupDP = (KnxGroupDP) container.getChannelHandle();
				}

				groupDP.getKnxValue().setOpenMucValue(container.getValue());
				boolean state = knxConnection.write(groupDP, timeout);
				if (state) {
					container.setFlag(Flag.VALID);
				}
				else {
					container.setFlag(Flag.UNKNOWN_ERROR);
				}
			} catch (KNXException e) {
				logger.warn(e.getMessage());
			}
		}

		return null;
	}

	private static KnxGroupDP createKnxGroupDP(String channelAddress) throws KNXException {
		String[] address = channelAddress.split(":");
		GroupAddress main = new GroupAddress(address[0]);
		String dptID = address[1];
		int mainNumber = Integer.parseInt(dptID.split("\\.")[0]);
		return new KnxGroupDP(main, channelAddress, mainNumber, dptID);
	}
}
