/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.driver.modbustcp;

import java.net.InetAddress;
import java.net.UnknownHostException;

import net.wimpi.modbus.ModbusException;
import net.wimpi.modbus.ModbusIOException;
import net.wimpi.modbus.ModbusSlaveException;
import net.wimpi.modbus.io.ModbusTCPTransaction;
import net.wimpi.modbus.msg.ReadCoilsRequest;
import net.wimpi.modbus.msg.ReadCoilsResponse;
import net.wimpi.modbus.msg.ReadInputDiscretesRequest;
import net.wimpi.modbus.msg.ReadInputDiscretesResponse;
import net.wimpi.modbus.msg.ReadInputRegistersRequest;
import net.wimpi.modbus.msg.ReadInputRegistersResponse;
import net.wimpi.modbus.msg.ReadMultipleRegistersRequest;
import net.wimpi.modbus.msg.ReadMultipleRegistersResponse;
import net.wimpi.modbus.msg.WriteCoilRequest;
import net.wimpi.modbus.msg.WriteMultipleCoilsRequest;
import net.wimpi.modbus.msg.WriteMultipleRegistersRequest;
import net.wimpi.modbus.msg.WriteSingleRegisterRequest;
import net.wimpi.modbus.net.TCPMasterConnection;
import net.wimpi.modbus.procimg.InputRegister;
import net.wimpi.modbus.procimg.Register;
import net.wimpi.modbus.util.BitVector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * 
 * TODO
 * 
 * @author Marco Mittelsdorf
 */
public class ModbusTcpConnection {

	private final static Logger logger = LoggerFactory.getLogger(ModbusTcpConnection.class);

	private InetAddress slaveAddress;
	private TCPMasterConnection connection;
	private ModbusTCPTransaction transaction;
	private ReadCoilsRequest readCoilsRequest;
	private ReadInputDiscretesRequest readInputDiscretesRequest;
	private WriteCoilRequest writeCoilRequest;
	private WriteMultipleCoilsRequest writeMultipleCoilsRequest;
	private ReadInputRegistersRequest readInputRegistersRequest;
	private ReadMultipleRegistersRequest readHoldingRegisterRequest;
	private WriteSingleRegisterRequest writeSingleRegisterRequest;
	private WriteMultipleRegistersRequest writeMultipleRegistersRequest;

	public ModbusTcpConnection(String addr) {
		try {
			slaveAddress = InetAddress.getByName(addr);
			connection = new TCPMasterConnection(slaveAddress);
			readCoilsRequest = new ReadCoilsRequest();
			readInputDiscretesRequest = new ReadInputDiscretesRequest();
			readInputRegistersRequest = new ReadInputRegistersRequest();
			readHoldingRegisterRequest = new ReadMultipleRegistersRequest();
			writeCoilRequest = new WriteCoilRequest();
			writeMultipleCoilsRequest = new WriteMultipleCoilsRequest();
			writeSingleRegisterRequest = new WriteSingleRegisterRequest();
			writeMultipleRegistersRequest = new WriteMultipleRegistersRequest();
		} catch (UnknownHostException e) {
			throw new RuntimeException(e.getMessage());
		}
	}

	public ModbusTcpConnection(String addr, int port) {
		this(addr);
		connection.setPort(port);
	}

	public void connect() throws Exception {
		if (connection != null && !connection.isConnected()) {
			connection.connect();
			transaction = new ModbusTCPTransaction(connection);
		}
	}

	public void disconnect() {
		if (connection != null && connection.isConnected()) {
			connection.close();
			transaction = null;
		}
	}

	public void setTimeout(int timeout) {
		connection.setTimeout(timeout);
	}

	private synchronized BitVector readCoils(int startAddress, int count, int unitID) throws ModbusException {
		readCoilsRequest.setReference(startAddress);
		readCoilsRequest.setBitCount(count);
		readCoilsRequest.setUnitID(unitID);
		transaction.setRequest(readCoilsRequest);
		transaction.execute();
		return ((ReadCoilsResponse) transaction.getResponse()).getCoils();
	}

	public BitVector readCoils(ModbusTcpChannel channel) throws ModbusException {
		return readCoils(channel.getStartAddress(), channel.getCount(), channel.getUnitId());
	}

	public BitVector readCoils(ModbusTcpChannelGroup channelGroup) throws ModbusException {
		return readCoils(channelGroup.getStartAddress(), channelGroup.getCount(), channelGroup.getUnitId());
	}

	private synchronized BitVector readDiscreteInputs(int startAddress, int count, int unitID) throws ModbusException {
		readInputDiscretesRequest.setReference(startAddress);
		readInputDiscretesRequest.setBitCount(count);
		readInputDiscretesRequest.setUnitID(unitID);
		transaction.setRequest(readInputDiscretesRequest);
		transaction.execute();
		return ((ReadInputDiscretesResponse) transaction.getResponse()).getDiscretes();
	}

	public BitVector readDiscreteInputs(ModbusTcpChannel channel) throws ModbusException {
		return readDiscreteInputs(channel.getStartAddress(), channel.getCount(), channel.getUnitId());
	}

	public BitVector readDiscreteInputs(ModbusTcpChannelGroup channelGroup) throws ModbusException {
		return readDiscreteInputs(channelGroup.getStartAddress(), channelGroup.getCount(), channelGroup.getUnitId());
	}

	private synchronized Register[] readHoldingRegisters(int startAddress, int count, int unitID)
			throws ModbusException {
		readHoldingRegisterRequest.setReference(startAddress);
		readHoldingRegisterRequest.setWordCount(count);
		readHoldingRegisterRequest.setUnitID(unitID);
		transaction.setRequest(readHoldingRegisterRequest);
		transaction.execute();
		return ((ReadMultipleRegistersResponse) transaction.getResponse()).getRegisters();
	}

	public Register[] readHoldingRegisters(ModbusTcpChannel channel) throws ModbusException {
		return readHoldingRegisters(channel.getStartAddress(), channel.getCount(), channel.getUnitId());
	}

	public Register[] readHoldingRegisters(ModbusTcpChannelGroup channelGroup) throws ModbusException {
		return readHoldingRegisters(channelGroup.getStartAddress(), channelGroup.getCount(), channelGroup.getUnitId());
	}

	/**
	 * Read InputRegisters
	 * 
	 * @param startAddress
	 * @param count
	 * @param unitID
	 * @return the InputRegister[]
	 * @throws ModbusIOException
	 * @throws ModbusSlaveException
	 * @throws ModbusException
	 */
	private synchronized InputRegister[] readInputRegisters(int startAddress, int count, int unitID)
			throws ModbusIOException, ModbusSlaveException, ModbusException {
		readInputRegistersRequest.setReference(startAddress);
		readInputRegistersRequest.setWordCount(count);
		readInputRegistersRequest.setUnitID(unitID);
		transaction.setRequest(readInputRegistersRequest);
		transaction.execute();
		InputRegister[] registers = ((ReadInputRegistersResponse) transaction.getResponse()).getRegisters();
		return registers;
	}

	/**
	 * Read InputRegisters for a channel
	 * 
	 * @param channel
	 * @return the InputRegister
	 * @throws ModbusException
	 */
	public InputRegister[] readInputRegisters(ModbusTcpChannel channel) throws ModbusException {
		return readInputRegisters(channel.getStartAddress(), channel.getCount(), channel.getUnitId());
	}

	/**
	 * Read InputRegisters for a channelGroup
	 * 
	 * @param channelGroup
	 * @return the InputRegister
	 * @throws ModbusException
	 */
	public InputRegister[] readInputRegisters(ModbusTcpChannelGroup channelGroup) throws ModbusException {
		return readInputRegisters(channelGroup.getStartAddress(), channelGroup.getCount(), channelGroup.getUnitId());
	}

	public synchronized void writeSingleCoil(ModbusTcpChannel channel, boolean state) throws ModbusException {
		writeCoilRequest.setReference(channel.getStartAddress());
		writeCoilRequest.setCoil(state);
		writeCoilRequest.setUnitID(channel.getUnitId());
		transaction.setRequest(writeCoilRequest);
		transaction.execute();
	}

	public synchronized void writeMultipleCoils(ModbusTcpChannel channel, BitVector coils) throws ModbusException {
		writeMultipleCoilsRequest.setReference(channel.getStartAddress());
		writeMultipleCoilsRequest.setCoils(coils);
		writeMultipleCoilsRequest.setUnitID(channel.getUnitId());
		transaction.setRequest(writeMultipleCoilsRequest);
		transaction.execute();
	}

	public synchronized void writeSingleRegister(ModbusTcpChannel channel, Register register) throws ModbusException {

		writeSingleRegisterRequest.setReference(channel.getStartAddress());
		writeSingleRegisterRequest.setRegister(register);
		writeSingleRegisterRequest.setUnitID(channel.getUnitId());
		transaction.setRequest(writeSingleRegisterRequest);
		transaction.execute();
	}

	public synchronized void writeMultipleRegisters(ModbusTcpChannel channel, Register[] registers)
			throws ModbusException {
		writeMultipleRegistersRequest.setReference(channel.getStartAddress());
		writeMultipleRegistersRequest.setRegisters(registers);
		writeMultipleRegistersRequest.setUnitID(channel.getUnitId());
		transaction.setRequest(writeMultipleRegistersRequest);
		transaction.execute();
	}

}
