/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.driver.modbustcp;

import java.util.List;

import net.wimpi.modbus.Modbus;
import net.wimpi.modbus.ModbusException;
import net.wimpi.modbus.procimg.InputRegister;
import net.wimpi.modbus.procimg.Register;
import net.wimpi.modbus.procimg.SimpleRegister;
import net.wimpi.modbus.util.BitVector;

import org.openmuc.framework.data.Flag;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.data.Value;
import org.openmuc.framework.driver.spi.ChannelRecordContainer;
import org.openmuc.framework.driver.spi.ConnectionException;

public class ModbusTcpDevice {

	private ModbusTcpConnection connection = null;
	private String slaveAddress;
	private int port;
	ModbusTcpUtil util;

	/**
	 * 
	 * @param deviceAddress
	 *            e.g. ipadress, ipadress:port
	 */
	public ModbusTcpDevice(String deviceAddress) {
		setAddress(deviceAddress);

		connection = new ModbusTcpConnection(slaveAddress, port);
		util = new ModbusTcpUtil();
	}

	public void connect() throws ConnectionException {
		try {
			connection.connect();
		} catch (Exception e) {
			throw new ConnectionException("Unable to open Modbus connection for device with address " + slaveAddress
					+ " and port " + port);
		}
	}

	public void disconnect() {
		connection.disconnect();
	}

	/**
	 * 
	 * @param deviceAddress
	 *            e.g. localhost:1502
	 */
	private void setAddress(String deviceAddress) {
		String[] address = deviceAddress.toLowerCase().split(":");

		slaveAddress = null;
		port = Modbus.DEFAULT_PORT;

		if (address.length == 1) {
			slaveAddress = address[0];
		}
		else if (address.length == 2) {
			slaveAddress = address[0];
			port = Integer.parseInt(address[1]);
		}
		else {
			throw new RuntimeException("Invalid device address: '" + deviceAddress
					+ "'! Use following format: [ip:port] like localhost:1502 or 127.0.0.1:1502");
		}
	}

	/**
	 * Read a single channel
	 * 
	 * @param channel
	 * @return a single Value for the ModbusTcpChannel
	 * @throws Exception
	 */
	public Value readChannel(ModbusTcpChannel channel, int timeout) throws ModbusException {
		Value value = null;

		connection.setTimeout(timeout);

		switch (channel.getFunctionCode()) {
		case FC_01_READ_COILS:
			value = util.getBitVectorsValue(connection.readCoils(channel));
			break;
		case FC_02_READ_DISCRETE_INPUTS:
			value = util.getBitVectorsValue(connection.readDiscreteInputs(channel));
			break;
		case FC_03_READ_HOLDING_REGISTERS:
			value = util.getRegistersValue(connection.readHoldingRegisters(channel), channel.getDatatype());
			break;
		case FC_04_READ_INPUT_REGISTERS:
			value = util.getRegistersValue(connection.readInputRegisters(channel), channel.getDatatype());
			break;
		default:
			throw new RuntimeException("FunctionCode " + channel.getFunctionCode() + " not supported yet");
		}

		return value;
	}

	/**
	 * Read a group of channels
	 * 
	 * @param channelGroup
	 * @throws ModbusException
	 */
	public void readChannelGroup(ModbusTcpChannelGroup channelGroup, List<ChannelRecordContainer> containers,
			int timeout) throws ModbusException {

		connection.setTimeout(timeout);

		switch (channelGroup.getFunctionCode()) {
		case FC_01_READ_COILS:
			BitVector coils = connection.readCoils(channelGroup);
			channelGroup.setChannelValues(coils, this, containers);
			break;
		case FC_02_READ_DISCRETE_INPUTS:
			BitVector discretInput = connection.readDiscreteInputs(channelGroup);
			channelGroup.setChannelValues(discretInput, this, containers);
			break;
		case FC_03_READ_HOLDING_REGISTERS:
			Register[] registers = connection.readHoldingRegisters(channelGroup);
			channelGroup.setChannelValues(registers, this, containers);
			break;
		case FC_04_READ_INPUT_REGISTERS:
			InputRegister[] inputRegisters = connection.readInputRegisters(channelGroup);
			channelGroup.setChannelValues(inputRegisters, this, containers);
			break;
		default:
			throw new RuntimeException("FunctionCode " + channelGroup.getFunctionCode() + " not supported yet");
		}
	}

	public void writeChannel(ModbusTcpChannel channel, Value value) throws ModbusException, RuntimeException {

		switch (channel.getFunctionCode()) {
		case FC_05_WRITE_SINGLE_COIL:
			connection.writeSingleCoil(channel, value.asBoolean());
			break;
		case FC_15_WRITE_MULITPLE_COILS:
			connection.writeMultipleCoils(channel, util.getBitVectorFromByteArray(value));
			break;
		case FC_06_WRITE_SINGLE_REGISTER:
			connection.writeSingleRegister(channel, new SimpleRegister(value.asShort()));
			break;
		case FC_16_WRITE_MULTIPLE_REGISTERS:
			connection.writeMultipleRegisters(channel, util.valueToRegisters(value, channel.getDatatype()));
			break;
		default:
			throw new RuntimeException("FunctionCode " + channel.getFunctionCode().toString() + " not supported yet");
		}
	}

	public void setChannelsWithErrorFlag(List<ChannelRecordContainer> containers) {
		for (ChannelRecordContainer container : containers) {
			container.setRecord(new Record(null, null, Flag.DRIVER_ERROR_CHANNEL_TEMPORARILY_NOT_ACCESSIBLE));
		}
	}

}
