/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.webui.base;

import java.util.Hashtable;
import java.util.Set;

import org.apache.velocity.app.Velocity;
import org.openmuc.framework.authentication.AuthenticationService;
import org.openmuc.framework.webui.spi.ResourceLoader;
import org.openmuc.framework.webui.spi.WebUiPluginService;
import org.osgi.service.component.ComponentContext;
import org.osgi.service.http.HttpService;
import org.osgi.service.http.NamespaceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class WebUiBase {

	private final static Logger logger = LoggerFactory.getLogger(WebUiBase.class);

	private final Hashtable<String, WebUiPluginService> pluginsByAlias = new Hashtable<String, WebUiPluginService>();
	private HttpService httpService;
	private AuthenticationService authService;

	protected void activate(ComponentContext context) throws Exception {
		logger.info("Activating WebUI Base");

		Velocity.setProperty("resource.loader", "class");
		Velocity.setProperty("class.resource.loader.class",
				"org.apache.velocity.runtime.resource.loader.ClasspathResourceLoader");

		Velocity.init();

		httpService.registerResources("/openmuc/css", "/css", null);
		httpService.registerResources("/openmuc/images", "/images", null);
		httpService.registerResources("/openmuc/js", "/js", null);
		httpService.registerServlet("/openmuc", new WebUiBaseServlet(new ResourceLoader(context.getBundleContext()),
				this), null, null);

	}

	protected void deactivate(ComponentContext context) {
		logger.info("Deactivating WebUI Base");
		httpService.unregister("/openmuc");
		httpService.unregister("/openmuc/css");
		httpService.unregister("/openmuc/images");
		httpService.unregister("/openmuc/js");
	}

	protected void setHttpService(HttpService httpService) {
		this.httpService = httpService;
	}

	protected void unsetHttpService(HttpService httpService) {
		this.httpService = null;
	}

	protected void setWebUiPluginService(WebUiPluginService uiPlugin) {
		logger.info("Adding WebUI plugin: " + uiPlugin.getName());

		pluginsByAlias.put(uiPlugin.getAlias(), uiPlugin);
		registerResources(uiPlugin);
	}

	protected void unsetWebUiPluginService(WebUiPluginService uiPlugin) {
		logger.info("Removing WebUI plugin: " + uiPlugin.getName());

		unregisterResources(uiPlugin);
		pluginsByAlias.remove(uiPlugin.getAlias());
	}

	protected void setAuthenticationService(AuthenticationService authService) {
		this.authService = authService;
	}

	protected void unsetAuthenticationService(AuthenticationService authService) {
		this.authService = null;
	}

	private void registerResources(WebUiPluginService plugin) {
		Set<String> aliases = plugin.getResources().keySet();

		for (String alias : aliases) {
			try {

				httpService.registerResources("/openmuc/" + plugin.getCategory() + "/" + plugin.getAlias() + "/"
						+ alias, plugin.getResources().get(alias), plugin);

			} catch (NamespaceException e) {
				e.printStackTrace();
			}
		}
	}

	private void unregisterResources(WebUiPluginService plugin) {
		Set<String> aliases = plugin.getResources().keySet();

		for (String alias : aliases) {
			httpService.unregister("/openmuc/" + plugin.getCategory().toString() + "/" + plugin.getAlias() + "/"
					+ alias);
		}
	}

	public WebUiPluginService getPlugin(String alias) {
		return pluginsByAlias.get(alias);
	}

	public Hashtable<String, WebUiPluginService> getPlugins() {
		return pluginsByAlias;
	}

	public AuthenticationService getAuthenticationService() {
		return authService;
	}
}
