/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.webui.channelconfigurator;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Hashtable;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.openmuc.framework.config.ChannelConfig;
import org.openmuc.framework.config.ConfigService;
import org.openmuc.framework.config.ConfigWriteException;
import org.openmuc.framework.config.DeviceConfig;
import org.openmuc.framework.config.DriverConfig;
import org.openmuc.framework.config.NamingCollisionException;
import org.openmuc.framework.config.RootConfig;
import org.openmuc.framework.data.ValueType;
import org.openmuc.framework.webui.spi.PluginCategory;
import org.openmuc.framework.webui.spi.PluginContext;
import org.openmuc.framework.webui.spi.View;
import org.openmuc.framework.webui.spi.WebUiPluginService;
import org.osgi.framework.BundleContext;
import org.osgi.service.component.ComponentContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class ChannelConfigurator implements WebUiPluginService {

	private final static Logger logger = LoggerFactory.getLogger(ChannelConfigurator.class);

	private BundleContext context;
	private ConfigService configService;
	private RootConfig rootConfig;
	private WebconfigViewFactory factory;

	protected void activate(ComponentContext context) {
		logger.info("Activating WebUI Channel Configurator");
		this.context = context.getBundleContext();
		factory = new WebconfigViewFactory(context);
	}

	protected void deactivate(ComponentContext context) {
		logger.info("Deactivating WebUI Channel Configurator");
	}

	protected void setConfigService(ConfigService configService) {
		this.configService = configService;
	}

	protected void unsetConfigService(ConfigService configService) {
		this.configService = null;
	}

	@Override
	public Hashtable<String, String> getResources() {
		Hashtable<String, String> resources = new Hashtable<String, String>();

		resources.put("css", "css");

		return resources;
	}

	@Override
	public String getAlias() {
		return "channelconfigurator";
	}

	@Override
	public String getName() {
		return "Channel Configurator";
	}

	@Override
	public String getDescription() {
		return "Configuration utility for the devices and channels.";
	}

	@Override
	public View getContentView(HttpServletRequest request, PluginContext context) {
		rootConfig = configService.getConfig();

		WebconfigView view = null;
		try {
			if (context.getLocalPath().equals("/resetconfig")) {
				try {
					configService.reloadConfigFromFile();

					view = factory.createMessage("Message", "Configuration reset from persistant storage");
				} catch (Exception e) {
					view = factory.createMessage("Error",
							"Error on reading configuration from persistant storage. Reason: " + e.getMessage());
				}
			}
			else if (context.getLocalPath().equals("/saveconfig")) {
				try {
					configService.writeConfigToFile();

					view = factory.createMessage("Message", "Configurations written to persistant storage");
				} catch (ConfigWriteException e) {
					view = factory.createMessage("Error", "Error on writing configuration. Reason: " + e.getMessage());
				}
			}
			else if (context.getLocalPath().equals("/fileoptions")) {
				view = factory.createFileOptions();
			}
			else if (context.getLocalPath().equals("/deletechannel")) {
				String driverId = readParameter(request, "driverId");
				String interfaceAddress = request.getParameter("interfaceAddress");
				String deviceAddress = readParameter(request, "deviceAddress");
				String channelAddress = readParameter(request, "channelAddress");

				DriverConfig driver = rootConfig.getDriver(driverId);
				DeviceConfig device = driver.getDevice(interfaceAddress, deviceAddress);
				ChannelConfig channel = device.getChannel(channelAddress);

				channel.delete();
				configService.setConfig(rootConfig);

				Collection<DriverConfig> driverList = rootConfig.getDrivers();

				view = factory.createChannelsView(driverList);
			}
			else if (context.getLocalPath().equals("/togglechannel")) {
				String driverId = readParameter(request, "driverId");
				String interfaceAddress = request.getParameter("interfaceAddress");
				String deviceAddress = readParameter(request, "deviceAddress");
				String address = readParameter(request, "address");

				DriverConfig driver = rootConfig.getDriver(driverId);
				DeviceConfig device = driver.getDevice(interfaceAddress, deviceAddress);
				ChannelConfig channel = device.getChannel(address);

				channel.setDisabled(!channel.isDisabled());
				configService.setConfig(rootConfig);

				view = factory.createChannelsView(rootConfig.getDrivers());
			}
			else if (context.getLocalPath().equals("/updatechannel")) {
				String driverId = readParameter(request, "driverId");
				String interfaceAddress = request.getParameter("interfaceAddress");
				String deviceAddress = readParameter(request, "deviceAddress");
				String address = readParameter(request, "address");

				String label = request.getParameter("label");
				String oldAddress = request.getParameter("oldAddress");

				DriverConfig driver = rootConfig.getDriver(driverId);
				DeviceConfig device = driver.getDevice(interfaceAddress, deviceAddress);

				ChannelConfig channel = null;
				if (oldAddress == null) {
					channel = device.addChannel(address);
					if (label.isEmpty() == false) {
						channel.setLabel(label);
					}
				}
				else {
					channel = device.getChannel(oldAddress);
					if (address.equals(channel.getChannelAddress()) == false) {
						channel.setChannelAddress(address);
					}
					if (label.equals(channel.getLabel()) == false) {
						channel.setLabel(label);
					}
				}

				List<String> rights = new ArrayList<String>();
				if (request.getParameterValues("accessRights") != null) {
					rights = Arrays.asList(request.getParameterValues("accessRights"));
				}

				channel.setDescription(request.getParameter("description"));
				channel.setValueType(ValueType.valueOf(request.getParameter("valueType")));
				channel.setValueTypeLength(tryParseInt(request.getParameter("valueLength")));
				channel.setScalingFactor(tryParseDouble(request.getParameter("scalingFactor")));
				channel.setValueOffset(tryParseDouble(request.getParameter("scalingOffset")));
				channel.setUnit(request.getParameter("unit"));
				channel.setLoggingInterval(tryParseInt(request.getParameter("loggingInterval")));
				channel.setLoggingTimeOffset(tryParseInt(request.getParameter("loggingOffset")));
				channel.setSamplingInterval(tryParseInt(request.getParameter("samplingInterval")));
				channel.setSamplingTimeOffset(tryParseInt(request.getParameter("samplingOffset")));
				channel.setSamplingGroup(request.getParameter("samplingGroup"));
				channel.setListening(request.getParameter("listening") != null);
				channel.setReadable(rights.contains("read"));
				channel.setWritable(rights.contains("write"));
				channel.setDisabled(request.getParameter("activity") != null);

				configService.setConfig(rootConfig);

				view = factory.createChannelsView(rootConfig.getDrivers());
			}
			else if (context.getLocalPath().equals("/editchannel")) {
				String driverId = readParameter(request, "driverId");
				String interfaceAddress = request.getParameter("interfaceAddress");
				String deviceAddress = readParameter(request, "deviceAddress");
				String channelAddress = request.getParameter("channelAddress");

				DriverConfig driver = rootConfig.getDriver(driverId);
				DeviceConfig device = driver.getDevice(interfaceAddress, deviceAddress);

				ChannelConfig channel = null;
				String title = "Add channel to " + device.getName();
				if (channelAddress != null && channelAddress.isEmpty() == false) {
					channel = device.getChannel(channelAddress);
					title = "Edit channel " + channel.getLabel();
				}

				view = factory.createEditChannelView(title, driverId, device, channel);
			}
			else if (context.getLocalPath().equals("/showchannels")) {
				String driverId = readParameter(request, "driverId");
				String interfaceAddress = readParameter(request, "interfaceAddress");
				String deviceAddress = readParameter(request, "deviceAddress");

				DriverConfig driver = rootConfig.getDriver(driverId);
				DeviceConfig device = driver.getDevice(interfaceAddress, deviceAddress);

				view = factory.createChannelsView(driverId, device);
			}
			else if (context.getLocalPath().equals("/channellist")) {
				view = factory.createChannelsView(rootConfig.getDrivers());
			}
			else if (context.getLocalPath().equals("/confirmdeletedevice")) {
				String driverId = readParameter(request, "driverId");
				String interfaceAddress = readParameter(request, "interfaceAddress");
				String deviceAddress = readParameter(request, "deviceAddress");
				String decision = readParameter(request, "decision");

				if ("delete".equals(decision)) {
					DriverConfig driver = rootConfig.getDriver(driverId);
					DeviceConfig device = driver.getDevice(interfaceAddress, deviceAddress);
					device.delete();
					configService.setConfig(rootConfig);
				}

				view = factory.createDeviceView(rootConfig.getDrivers());
			}
			else if (context.getLocalPath().equals("/deletedevice")) {
				String driverId = readParameter(request, "driverId");
				String interfaceAddress = request.getParameter("interfaceAddress");
				String deviceAddress = readParameter(request, "deviceAddress");

				DriverConfig driver = rootConfig.getDriver(driverId);
				DeviceConfig device = driver.getDevice(interfaceAddress, deviceAddress);

				view = factory.createDeleteDeviceView(driverId, device);
			}
			else if (context.getLocalPath().equals("/toggledevice")) {
				String driverId = readParameter(request, "driverId");
				String interfaceAddress = request.getParameter("interfaceAddress");
				String deviceAddress = readParameter(request, "deviceAddress");

				DriverConfig driver = rootConfig.getDriver(driverId);
				DeviceConfig device = driver.getDevice(interfaceAddress, deviceAddress);

				device.setDisabled(!device.isDisabled());
				configService.setConfig(rootConfig);

				view = factory.createDeviceView(rootConfig.getDrivers());
			}
			else if (context.getLocalPath().equals("/updatedevice")) {
				String driverId = readParameter(request, "driver");
				String interfaceAddress = request.getParameter("interfaceAddress");
				String deviceAddress = readParameter(request, "deviceAddress");
				String originalInterfaceAddress = request.getParameter("originalInterfaceAddress");
				String originalDeviceAddress = request.getParameter("originalDeviceAddress");

				DriverConfig driver = rootConfig.getDriver(driverId);
				DeviceConfig newDevice = null;
				if (originalDeviceAddress == null && originalInterfaceAddress == null) {
					newDevice = driver.addDevice(interfaceAddress, deviceAddress);
				}
				else {
					newDevice = driver.getDevice(originalInterfaceAddress, originalDeviceAddress);
					if (originalDeviceAddress.equals(deviceAddress) == false
							|| originalInterfaceAddress.equals(interfaceAddress) == false) {
						newDevice.setAddresses(interfaceAddress, deviceAddress);
					}
				}

				String activity = request.getParameter("activity");

				newDevice.setName(request.getParameter("name"));
				newDevice.setDescription(request.getParameter("description"));
				newDevice.setReadTimeout(tryParseInt(request.getParameter("readTimeout")));
				newDevice.setConnectTimeout(tryParseInt(request.getParameter("connectTimeout")));
				newDevice.setConnectRetryInterval(tryParseInt(request.getParameter("connectRetry")));
				newDevice.setSettings(request.getParameter("settings"));
				newDevice.setDisabled(activity != null && activity.equals("disable"));

				configService.setConfig(rootConfig);

				Collection<DriverConfig> driverList = rootConfig.getDrivers();

				view = factory.createDeviceView(driverList);
			}
			else if (context.getLocalPath().equals("/editdevice")) {
				String driverId = request.getParameter("driverId");
				String interfaceAddress = request.getParameter("interfaceAddress");
				String deviceAddress = request.getParameter("deviceAddress");

				Collection<DriverConfig> driverList = rootConfig.getDrivers();
				DeviceConfig device = null;
				String title = "Add new device";

				if (deviceAddress != null && deviceAddress.isEmpty() == false) {
					DriverConfig driver = rootConfig.getDriver(driverId);
					device = driver.getDevice(interfaceAddress, deviceAddress);
					driverList = Arrays.asList(driver);
					title = "Edit device " + device.getName();
				}

				view = factory.createEditDeviceView(title, driverList, device);
			}
			else if (context.getLocalPath().equals("/showdevices")) {
				String driverId = readParameter(request, "id");

				DriverConfig driver = rootConfig.getDriver(driverId);

				view = factory.createDeviceView(Arrays.asList(driver), "Devices of " + driver.getId());
			}
			else if (context.getLocalPath().equals("/devicelist")) {
				view = factory.createDeviceView(rootConfig.getDrivers());
			}
			else if (context.getLocalPath().equals("/confirmdeletedriver")) {
				String driverId = readParameter(request, "driverId");
				String decision = readParameter(request, "decision");

				if ("delete".equals(decision)) {
					rootConfig.getDriver(driverId).delete();
					configService.setConfig(rootConfig);
				}

				view = factory.createDriverView(rootConfig.getDrivers());
			}
			else if (context.getLocalPath().equals("/deletedriver")) {
				String driverId = readParameter(request, "id");

				DriverConfig driver = rootConfig.getDriver(driverId);

				view = factory.createDeleteDriverView(driver);
			}
			else if (context.getLocalPath().equals("/toggledriver")) {
				String driverId = readParameter(request, "id");

				DriverConfig driver = rootConfig.getDriver(driverId);

				driver.setDisabled(!driver.isDisabled());
				configService.setConfig(rootConfig);

				view = factory.createDriverView(rootConfig.getDrivers());
			}
			else if (context.getLocalPath().equals("/updatedriver")) {
				String oldId = request.getParameter("originalname");
				String driverId = readParameter(request, "name");

				DriverConfig newDriver = null;

				if (oldId == null) {
					newDriver = rootConfig.addDriver(driverId);
				}
				else {
					newDriver = rootConfig.getDriver(oldId);
					if (oldId.equals(driverId) == false) {
						newDriver.setId(driverId);
					}
				}

				String activity = request.getParameter("activity");

				newDriver.setReadTimeout(tryParseInt(request.getParameter("readTimeout")));
				newDriver.setConnectTimeout(tryParseInt(request.getParameter("connectTimeout")));
				newDriver.setConnectRetryInterval(tryParseInt(request.getParameter("connectRetry")));
				newDriver.setDisabled(activity != null && activity.equals("disable"));

				configService.setConfig(rootConfig);

				view = factory.createDriverView(rootConfig.getDrivers());
			}
			else if (context.getLocalPath().equals("/editdriver")) {
				String driverId = request.getParameter("id");

				DriverConfig driver = null;
				String title = "Add new driver";

				if (driverId != null && driverId.isEmpty() == false) {
					driver = rootConfig.getDriver(driverId);
					title = "Edit driver " + driver.getId();
				}

				view = factory.createEditDriverView(driver, title);
			}
			else {
				// Per default, show driver list
				view = factory.createDriverView(rootConfig.getDrivers());
			}
		} catch (ProcessRequestException pre) {
			view = factory.createMessage("Error", "Error on processing request. Please retry<br>Reason: " + pre);
		} catch (NullPointerException npe) {
			view = factory.createMessage("Error", "Error on processing request.");
		} catch (NamingCollisionException ce) {
			view = factory.createMessage("Error", ce.getMessage());
		}

		return view;
	}

	private Double tryParseDouble(String input) {
		try {
			return Double.parseDouble(input);
		} catch (NumberFormatException e) {
			return null;
		}
	}

	private Integer tryParseInt(String input) {
		try {
			return Integer.parseInt(input);
		} catch (NumberFormatException e) {
			return null;
		}
	}

	private String readParameter(HttpServletRequest request, String parameterName) throws ProcessRequestException {
		String result = request.getParameter(parameterName);

		if (result == null || result.isEmpty()) {
			throw new ProcessRequestException(parameterName + " not found in request");
		}

		return result;
	}

	@Override
	public boolean handleSecurity(HttpServletRequest request, HttpServletResponse response) throws IOException {
		return true;
	}

	@Override
	public URL getResource(String name) {
		return context.getBundle().getResource(name);
	}

	@Override
	public String getMimeType(String name) {
		return null;
	}

	@Override
	public PluginCategory getCategory() {
		return PluginCategory.CONFIGTOOL;
	}

}
